/*
 * samplestat.h :- collecting statistics; this file was extracted from
 * DaSSF.
 */

/*
 * Copyright (c) 1998-2002 Dartmouth College
 *
 * Permission is hereby granted, free of charge, to any individual or 
 * institution obtaining a copy of this software and associated 
 * documentation files (the "Software"), to use, copy, modify, and 
 * distribute without restriction, provided that this copyright and 
 * permission notice is maintained, intact, in all copies and supporting 
 * documentation.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL DARTMOUTH COLLEGE BE LIABLE FOR ANY CLAIM, DAMAGES 
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#ifndef __SAMPLESTAT_H__
#define __SAMPLESTAT_H__
#include <vector> 

using namespace std;

class SampleStat {
 private:
  vector<double> samples;

 public:
  /// Constructor
  SampleStat();

  /// Destructor
  virtual ~SampleStat();
  
  /// Record a new sample
  void record(double value); 

  /// Print summary table
  void report(FILE *outf);   

  /// Minimum of all samples
  double min();      

  /// Maximum of all samples
  double max();      

  /// Retrieve a sample based on an index that speaks to its order of recording
  double get_sample(int index);

  double range();    // i.e. max-min
  double size();     // sample size
  double mean();     // sample mean
  double variance(); // sample variance
  double stdev();    // sample standard deviation
  double var_coef(); // coefficient of variation, i.e. stdev/mean

  void set_conf_level(double conf = 0.95); // set confidence level
  double conf_level();           // return confidence level
  double conf_intv_half_width(); // half confidence interval width

  /*! Calculate the average of samples within a window.
      @param winpos position of window relative to the start of sample collection
      @param winsize number of samples in one window
   */
  double window_average(int winpos, int winsize);

  /*! Compute the moving averages over the sample collection. 

      The method creates, initializes, and returns an array of double; it
      will be the responsibility of the caller to free this memory to
      avoid memory leaks.

      @param winsize number of samples in one window @return an array
      of doubles containing all the window averages
   */
  vector<double> *moving_averages(int winsize);

protected:
  double _nentries;
  double _min, _max;
  double _mean, _sumsqr;
  double _conf_level;

  double student_idf(long n, double u);
};

class SampleHistogram
{
public:
  SampleHistogram(long nbuckets, double lo, double hi);
  SampleHistogram(long bufsiz = 1000);
  ~SampleHistogram();

  void record(double value);   // record a new sample
  void report(FILE *outf);     // print summary table

  long num_buckets();          // number of buckets
  double low();                // mid-point of lowest bucket
  double high();               // mid-point of highest bucket
  double width();              // bucket width
  double num_observations();   // total
  double num_outliers();       // underflow and overflow
  double frequency(long i);    // number of samples in bucket i
  double mean();               // histogram mean
  double variance();           // histogram variance
  double stdev();              // histogram stdev

protected:
  long _nbuckets;
  double _low, _high;
  double _width;
  double *_frequency;
  double _nentries;
  double _underliers, _overliers;
  int _mature;
  long _bufsiz;
  double *_buffer;
  int _mean_computed, _var_computed;
  double _mean, _var;

  void grow_up();
};

#endif /*__SAMPLESTAT_H__*/
