#ifndef MESSAGES_H
#define MESSAGES_H
/**
 * This header file contains C/C++ structs which mimic the structure found in
 * the python module safe.messages.  For more information see the original
 * python source.
 */
#define REGISTER 0
#define NEXT_REQUEST 1
#define NEXT_REPLY 2
#define SAMPLE 3
#define FINISHED  4
#define TERMINATE 5

namespace safe {

void SendSampleMessage(FILE* fp, struct sample_msg msg);

/**
 * This struct encapsulates the message that should be sent to the experiment
 * execution manager upon the connection being made.  This message should
 * communicate any relevant host specific configuration information to the
 * execution manager to save for future analysis.  This message should likely
 * only be sent from the simulation client to the execution manager.
 *
 * @param type should always be REGISTER
 * @info is the host specific information.  Additional fields may be added as
 * more relevant information is desired to be stored.
 */
struct register_msg {
	unsigned int type;
	char info[128];
};

/**
 * This message is sent by a simulation client to request the next design point
 * to simulate.  This message should only be transmitted from the simulation
 * client to the execution manager.
 *
 * @param type should always be NEXT_REQUEST
 */
struct next_request_msg {
	unsigned int type;
};

/**
 * This struct encapsulates the response from the experiment execution manager
 * to the simulation client informing it of the next design point.  It includes
 * the message type (NEXT_REPLY), the length of the xml configuration being
 * sent, and the xml itself.
 *
 * @param type should always be NEXT_REPLY
 * @param length is the length of the xml field
 * @param xml is the xml configuration of the design point to be simulated.
 */

struct next_reply_msg {
	unsigned int type;
	unsigned int length;
	char* xml; // one may need to be careful when serializing/encoding this.
};

/**
 * The SAMPLE message is the means for reporting simulation statistics samples to the execution manager.
 *
 * The simulator itself is encouraged to output statistics using this format to
 * minimize the amount of output analysis done on the part of the simulation
 * client.  If this format is used, the client can very easily compute the
 * transient and determine if it should forward the message to the execution
 * manager directly.
 *
 * @param type should always be SAMPLE
 * @param timestamp the simulated time at which point the sample was collected.
 * @param object a unique identifier of a simulated object to which the sample should be associated.
 * @param metric the type of sample the object is reporting.
 * @param value the value of the sample being reported.
 */

struct sample_msg {
	unsigned int type;
	double timestamp;
	char object[64];
	char metric[8];
	double value;
};

/**
 * The finished message indicates that no further samples for the active design
 * point are necessary.  This message should only be sent from the execution
 * manager to the simulation client.
 *
 * @param type should always be FINISHED
 */

struct finished_msg {
	unsigned int type;
};

/**
 * The termiate message indicates that no further samples are necessary for any
 * of the design points in the simulation.  No further requests for design
 * points will be serviced.  This message should signal the graceful
 * termination of the simulator and the simulation client itself.
 *
 * @param type should always be TERMINATE
 */

struct terminate_msg {
	unsigned int type;
};

}
#endif /*MESSAGES_H*/
