# -*- test-case-name: safe.test.models -*-
"""
The classes in the models module provide a layer of abstraction for data stored
in the relational database.  The DatabaseModel class provides this layer of
abstraction, and all other classes extend this functionality.
"""

from twisted.internet.defer import Deferred, DeferredList
from twisted.internet.threads import deferToThread
from twisted.internet import reactor, defer
from twisted.enterprise import adbapi
from datetime import datetime
from copy import deepcopy

from safe.exceptions import InvalidArgumentsException, DatabaseError
from safe.exceptions import ExecutionManagerError
from safe.xmltools.expparser import ExpParser
from safe.debug import *

class DatabaseModel(object):
    """DatabaseModel is an abstract class for database interactions.

    The DatabaseModel class provides a basic level of abstraction for a
    database table. An extending class will be able to insert, select, and
    update rows through a simple but intuitive syntax. All operations are
    designed to be run asynchronously in twisted.

    The DatabaseModel object provides some ORM like features, but it is by no
    means a full ORM. It does not provide any facilities for dealing with
    related tables through JOINs, and does not provide a query mechanism which
    can return multiple results. This class is meant to be a simple layer of
    abstraction for the database, as to date, the choice of ORMs in twisted is
    limited. This is due to problems with eager loading of data, which this
    implementation gets around by just not dealing with JOIN models.

    """

    def __init__(self, dbpool, **kwargs):
        """Default constructor for a database model."""

        self.dbpool = dbpool
        self.waiting = {}

        for k, v in kwargs.iteritems():
            ukey = "_"+k
            if k in self.columns:
                setattr(self, ukey, v)
            else:
                raise InvalidArgumentsException(
                        "Invalid keyword argument %s" % k)

    @classmethod
    def find(cls, dbpool, **kwargs):
        """Finder method for data model rows.

        This method is used to query for row object(s) of the specific class.
        If a single row is returned from the database query, the result will be
        returned a deferred to a single ``cls`` object.  In the case that
        multiple rows are returned a deferred to a list of ``cls`` objects will
        be returned.

        Keyword arguments for this method are ``cache`` and any column in the
        respective database table.  By default, the resulting query will only
        return the id, or primary key of the result, and any other columns need
        be queried for at a later time.  For efficiency, it is possible to
        include a list of strings of column names to store locally for faster
        access later.::

            Experiment.find(dbpool, id = id, cache=['name', template_id'])

        Keyword arguments which are database columns will then use the value of
        the keyword argument in the ``WHERE`` statement to filter search
        results. For example::

            Experiment.find(dbpool, template_id = 17)

        will return a list of experiment objects using template 17. These
        columns are *not* automatically included in the cached variables.
        """

        def createObj(results):
            """Callback for the SELECT query for find()."""
            if len(results) == 1:
                retval = cls(dbpool)
                for k, v in zip(cache,results[0]):
                    ukey = "_"+k
                    setattr(retval, ukey, v)

            elif len(results) > 1:
                retval = []
                for r in results:
                    tmp = cls(dbpool)
                    for k, v in zip(cache,r):
                        ukey = "_"+k
                        setattr(tmp, ukey, v)
                    retval.append(tmp)
            else:
                retval = None

            return retval

        cache = ['id']
        if kwargs.has_key('cache'):
            for c in kwargs['cache']:
                if c in cls.columns:
                    cache.append(c)
                else:
                    raise DatabaseError("Unknown column %s to cache" % c)
            del kwargs['cache']

        wheres = []
        for k, v in kwargs.iteritems():
            if k in cls.columns or k == 'id':
                if type(v) == str:
                    wheres.append("%s = $escape$%s$escape$" % (k,v))
                else:
                    wheres.append("%s = %s" % (k,str(v)))
            else:
                raise DatabaseError("Unknown column %s" % k)

        if len(wheres) > 0:
            query = """
            SELECT %s FROM %s WHERE %s
            """ % (','.join(cache), cls.table, ' AND '.join(wheres))
        else:
            query = """
            SELECT %s FROM %s
            """ % (','.join(cache), cls.table)

        d = dbpool.runQuery(query)
        d.addCallback(createObj)
        return d

    def _cacheResult(self, result, key):
        """Return the result of a query and save it as an attribute."""
        if len(result) == 1:
            retval = result[0][0]
            setattr(self, '_'+key, retval)
            
            for d in self.waiting[key]:
                d.callback(retval)

        else:
            raise DatabaseError("Relation not found")

    def _fetch_column(self, key):
        """Return a value of a column or attribute of the row.

        This method will always return a deferred.

        """
        ukey = "_"+key

        if key in self.columns and hasattr(self, ukey):
            return defer.succeed(getattr(self, ukey))
        elif key in self.columns and\
                hasattr(self, "_id") and getattr(self, "_id"):

            if not self.waiting.has_key(key):
                query = """
                SELECT %s FROM %s WHERE id = %d
                """ % (key, self.table, self._id)

                answer = Deferred()
                self.waiting[key] = [answer]

                d = self.dbpool.runQuery(query)
                d.addCallback(self._cacheResult, key)

                return answer

            else:
                d = Deferred()
                self.waiting[key].append(d)
                return d

        elif key in self.columns:
            return defer.succeed(None)
        elif key == "id" and hasattr(self, "_id") and getattr(self, "_id"):
            return defer.succeed(getattr(self, ukey))
        elif key == "id" and self.waiting.has_key('id'):
            d = Deferred()
            self.waiting['id'].append(d)
            return d
        else:
            return defer.fail(DatabaseError(
                "Relation has not been saved to obtain an id"))

    def __getitem__(self, key):
        """ Access cached column values through dictionary syntax.
        
        When accessing columns through an attribute such as
        ``experiment.name``, the value returned is via a deferred. This is to
        ensure that in the case that the value is not cached in memory, it will
        be fetched and returned to the deferred. Sometimes we know that a
        column value has been cached, in which case returning a deferred is a
        hassle to deal with. In such a case, the we access the cached value
        through dictionary or array like syntax such as experiment['name']
        which returns the cached value immediately. In the case that the value
        is not cached an ``KeyError`` will be raised.
        
        """

        if key in self.columns or key == 'id':
            ucol = "_"+key
            if hasattr(self, ucol):
                return getattr(self, ucol)
            else:
                raise KeyError("Column %s not currently cached" % key)
        else:
            raise KeyError("Unknown key %s" % column)

    def __getattr__(self, name):
        """Method invoked when an attribute is requested and not found.
       
        All cached database columns are stored as self._column.  This method is
        invoked when an attribute is not found, and if it is a column, then we
        fetch the value of that database column.
        
        """
        
        if name == "table":
            return self.__class__.__name__.lower()+"s"
        elif name in self.columns or name == "id":
            return self._fetch_column(name)
        else:
            raise AttributeError("Object has no attribute '%s'" % name)

    def __setattr__(self, name, value):
        """Method invoked when an attribute of the method is set.

        One must take particular care in this case because unlike __getattr__,
        this method is invoked when any attribute is set, not just when the
        default set method fails.  This method must therefor set any attributes
        normally, in addition to handling the database columns.

        """

        if name[0] == "_" and name[1:] in self.columns:
            self.__dict__[name] = value
            self._dirty = True
        elif name in self.columns:
            self.__dict__["_"+name] = value
            self._dirty = True
        else:
            self.__dict__[name] = value

    def __setitem__(self, name, value):
        """This method allows one to update the row using dict syntax.

        When the name of a column is given with syntax such as

        row["name"] = value

        then this method will update self._name = value, as well as indicate
        that a local change has been made, so the local copy is dirty relative
        to the database.

        """

        uname = "_"+name
        if name in self.columns and\
                (not hasattr(self,uname) or getattr(self,uname) != value):

            self.__dict__[uname] = value
            self._dirty = True


    def save(self):
        """Save the new row to the database.

        If the row is determined to already be in the database, save
        should not be executed.  To save any changes to the database, use
        update instead of save.

        """

        def respond(result):
            if len(result) == 1:
                retval = result[0][0]
                self._id = retval
                self._dirty = False

                for d in self.waiting['id']:
                    d.callback(retval)

                del self.waiting['id']

                return retval

            else:
                raise DatabaseError("Relation not found")

        if not self.waiting.has_key('id'):
            cols = []
            vals = []
            for col in self.columns:
                ucol = "_"+col
                if hasattr(self, ucol):
                    cols.append(col)
                    val = getattr(self,ucol)
                    if type(val) == str:
                        val = "$escape$%s$escape$" % val
                    vals.append(val)

            cols = ",".join(cols)
            vals = ",".join([str(v) for v in vals])

            query = """
            INSERT INTO %s (%s) VALUES (%s) RETURNING id
            """ % (self.table, cols, vals)

            answer = Deferred()
            self.waiting['id'] = [answer]

            d = self.dbpool.runQuery(query)
            d.addCallback(respond)
            d.addErrback(printError)

            return answer
        else:
            return defer.fail(DatabaseError(
                '%s: save command already issued' % (type(self).__name__)))

    def update(self,**kwargs):
        """Method to update the database with local changes.

        This method must be careful to fail in the case that the row has
        not already been saved to the database.  Furthermore, for performance
        reasons, we also do not update the database unless we have local
        changes as determined by self._dirty.

        """
        if hasattr(self, "_id") and getattr(self, "_id"):
            for k,v in kwargs.iteritems():
                if k in self.columns:
                    setattr(self,"_"+k,v)
                else:
                    raise DatabaseError("%s not a valid column to update" % k)

            if self._dirty:
                updates = []
                for col in self.columns:
                    ucol = "_"+col
                    if hasattr(self, ucol):
                        val = getattr(self, ucol)
                        if type(val) == str:
                            val = "$escape$%s$escape$" % val
                        updates.append("%s = %s" % (col, val))

                query = """
                UPDATE %s SET %s WHERE id = %d
                """ % (self.table, ", ".join(updates), self._id)

                return self.dbpool.runOperation(query)
            else:
                return defer.succeed(True)
        else:
            return defer.fail(DatabaseError(
                "Can't update an unsaved row."))

    def delete(self, cascade = ""):
        """Remove the row from the database.
        
        It is important to delete a row from the database using row.delete()
        instead of del row, as if this method were __del__ instead, then every
        time the program terminated, it would call __del__ to deconstruct and
        recover memory before terminating, and thereby also delete any data in
        the database.
        
        """

        if cascade:
            cascade = "CASCADE"

        if hasattr(self,"_id") and getattr(self,"_id"):
            query = """
            DELETE FROM %s %s WHERE id = %d
            """ % (self.table, cascade, self._id)

            return self.dbpool.runOperation(query)

        return defer.succeed(True)

class Experiment(DatabaseModel):
    """Experiment is a data model for a simulation experiment.

    The Experiment class manages any interactions with an experiment, including
    but not limited to the creation of DesignPoints, Descriptions,
    and any database interactions with the experiments table.

    """
    
    table = 'experiments'
    columns = ['name', 'created_at', 'finished_at', 
            'description_id', 'doe_complete']

    def __init__(self,dbpool,file,**kwargs):

        self.ready = False
        self.file = file

        DatabaseModel.__init__(self,dbpool,**kwargs)

    def prepare_for_execution(self):
        """ Cache all of the variables needed for execution.

        This ensures that everything is available and does not need to be
        access through a deferred.

        """
        def respond_to_save(result = None):
            """ After the experiment is saved, get these things.

            N.B. this method is a callback and an errback, so if the experiment
            has already beeen saved, we can proceed and get these things.

            """

            dlist = DeferredList([self.id,
                    self.doe_complete,
                    self.description()]).addCallback(parse_xml)

        def parse_xml(((id_s, id),
            (complete_s, complete),
            (description_s, description))):

            if complete:
                self.ready = True
                self._doe_stack = []
                answer.callback(True)
            else:
                self.parse_configs().addCallback(setup_doe_stack)

        def setup_doe_stack(results):
            self._doe_stack = [
                    (self.factors.keys(), {}, deepcopy(self.factors))]
            self.ready = True
            answer.callback(True)

        answer = Deferred()
        if hasattr(self,"_id"):
            respond_to_save()
        else:
            self.save().addCallback(respond_to_save)
        return answer

    def description(self):
        """Get a Description object for the description_id."""
        def respond(id):
            Description.find(self.dbpool, id = id).addCallback(
                    answer.callback)

        answer = Deferred()
        self.description_id.addCallback(respond)
        return answer

    def next_design_point(self):
        """ Retrieve the next design point to dispatch to a simulation client.

        **Run prepare_for_execution before calling next_design_point initially.**

        This method will first compute all of the desing points, one by one as
        more design points are requested. After all design points have
        computed, then design points will be dispatched in order of the fewest
        number of samples collected.

        """


        if not self.ready:
            return defer.fail(
                    ExecutionManagerError("Execution manager not ready."))


        def new_constraints(self, factor, level, myFactorLevels):
            """Compute all of the new constraints after assigning factor a
            level.

            TODO: change this method to reflect dictionaries of restrictions
            instead of only 2-tuples.

            """

            constraints = []
            #for r in self.restrictions:
            #    if r[0][0] == factor and r[0][1] == level and\
            #        r[1][1] in myFactorLevels[r[1][0]]:

            #        constraints.append(r[1])
            #    elif r[1][0] == factor and r[1][1] == level and\
            #        r[0][1] in myFactorLevels[r[0][0]]:

            #        constraints.append(r[0])
            return constraints


        def respond_to_xml(xml):
            sim = DesignPoint(self.dbpool,
                    experiment_id = self._id,
                    config = xml,
                    completed = False)

            sim.save().addCallback(lambda _: answer.callback(sim))

        def apply_to_template(sim):
            xmld = deferToThread(run_in_thread, sim)
            xmld.addCallback(respond_to_xml)

        def run_in_thread(sim):
            designPoint = self.file
            for tuple in sim.items():
              designPoint = designPoint + " --" + tuple[0] + "=" + str(tuple[1])

            return designPoint

        answer = Deferred()
        
        if len(self._doe_stack) > 0:
            found = False
            while not found:
                keys, sim, myFactorLevels= self._doe_stack.pop()

                if len(keys) == 0:
                    apply_to_template(sim)
                    found = True
                else:
                    constraints = sorted(
                            [(len(myFactorLevels[x]),x) for x in keys])
                    count, factor = constraints.pop(0)
                    keys.remove(factor)

                    for level in myFactorLevels[factor]:
                        sim[factor] = level
                        new_cts = new_constraints(
                                self, factor, level, myFactorLevels) 

                        for f,l in new_cts:
                            myFactorLevels[f].remove(l)

                        self._doe_stack.append((deepcopy(keys), deepcopy(sim),
                            deepcopy(myFactorLevels)))

                        for f,l in new_cts:
                            myFactorLevels[f].append(l)

        else:

            def find_design_point(result):
                if len(result) == 1:
                    simd = DesignPoint.find(self.dbpool, id = result[0][0])
                    simd.addCallback(lambda sim: answer.callback(sim))
                    simd.addErrback(printError)
                else:
                    answer.callback(None)

            self.doe_complete = True
            self.update()

            query = """
                WITH mysims (id,config) AS (
                    SELECT  design_points.id,
                            design_points.config
                    FROM    design_points
                    WHERE   design_points.experiment_id = %d
                    AND     design_points.completed = FALSE
                )
                SELECT  sims.id,
                        sims.config,
                        samples.samples,
                        hosts.hosts
                FROM    mysims AS sims
                LEFT JOIN (
                    SELECT  sims.id,
                            count(*) AS samples
                    FROM    mysims AS sims
                    LEFT JOIN results
                    ON (sims.id = results.design_point_id)
                    GROUP BY sims.id
                ) AS samples
                    ON (sims.id = samples.id)
                LEFT JOIN (
                    SELECT  sims.id,
                            count(*) AS hosts
                    FROM    mysims AS sims
                    LEFT JOIN simulations AS hosts
                        ON (sims.id = hosts.design_point_id)
                    GROUP BY sims.id,hosts.design_point_id
                ) AS hosts
                    ON (sims.id = hosts.id)
                ORDER BY samples ASC, hosts ASC
                LIMIT 1
            """ % (self._id)

            d = self.dbpool.runQuery(query)
            d.addCallback(find_design_point)
            d.addErrback(printError)

        return answer

    def parse_configs(self):
        """Parse the experiment configuration files."""

        def respond(description):
            """Callback for getting the Description XML."""

            threaded = deferToThread(parse, description)
            threaded.addCallback(answer.callback)

        def parse(description):
            """Parse the input files."""

            parser = ExpParser(description)
            # Doesn't actually parse until we say so.
            parser.parse()

            self.factors = parser.get_factors_levels()
            self.restrictions = parser.get_restriction_expression()

            return (self.factors, self.restrictions)

        answer = Deferred()

        self.description().addCallback(
                lambda d: d.description).addCallback(respond)

        return answer

class DesignPoint(DatabaseModel):
    """A data model for a design point.

    This class can manage interactions with the design_points database table.
    
    """
    
    columns = ['experiment_id', 'config', 'completed', 'index']
    table = 'design_points'

    def check_completed(self):
        """Checks to see if enough samples have been collected."""

        def check(count):
            if count > 30:
                return True
            else:
                return False

        return self.samples_count().addCallback(check)

    def samples_count(self):
        """Return the number of samples collected for the design point."""

        def sanitize(result):
            if len(result) == 1:
                return result[0][0]
            else:
                return DatabaseError("Row not found.")

        if self._id:
            query = """
            SELECT count(*) FROM results WHERE design_point_id = %d
            """ % (self._id)

            return self.dbpool.runQuery(query).addCallback(sanitize)
        else:
            return defer.fail(
                    DatabaseError("DesignPoint has not been saved."))
    
class Host(DatabaseModel):
    """Host is a data model for a simulation client.

    This model keeps data specific on a simulation clients configuration.  This
    information is associated with simulations executed and samples collected.

    """

    table = 'hosts'
    columns = ['kernel', 'simulator']

class Result(DatabaseModel):
    """A data model for a statistical sample.

    This class can manage interactions with the results database table.
    
    """

    columns = ['design_point_id', 'host_id',
            'timestamp', 'object', 'metric', 'value']
    table = 'results'

class Simulation(DatabaseModel):
    """Simulation is a single simulation run for a design point.
    
    Simulation acts as a join model for hosts and simulations. This model
    allows for a a single DesignPoint to be associated with multiple hosts and
    multiple hosts to be associated with a single design point.

    """

    table = 'simulations'
    columns = ['host_id', 'design_point_id', 'started_at', 'finished_at',
            'transient', 'terminated', 'sha2']

class Description(DatabaseModel):
    """ Data model for an experiment description file.

    This class can manage interactions with the descriptions database table.
    
    """

    columns = ['description']
    table = 'descriptions'

