from twisted.trial import unittest

from safe.debug import *
from safe.boolean import BooleanTerm, Expression, Term, OrList, AndList

class BooleanTermTest(unittest.TestCase):
    """ Tests for the BooleanTerms class."""

    def setUp(self):
        self.t1 = Term("param1", 1)
        self.t2 = Term("param2", 1)
        self.t3 = Term("param3", 1)

    def testAnd(self):
        result = self.t1 & self.t2
        self.assertIsInstance(result, Expression)
        self.assertTrue(result.evaluate({"param1":1, "param2":1}))
        self.assertFalse(result.evaluate({"param1":1, "param2":0}))
        self.assertFalse(result.evaluate({"param1":0, "param2":1}))
        self.assertFalse(result.evaluate({"param1":0, "param2":0}))

    def testOr(self):
        result = self.t1 | self.t2
        self.assertIsInstance(result, Expression)
        self.assertTrue(result.evaluate({"param1":1, "param2":1}))
        self.assertTrue(result.evaluate({"param1":1, "param2":0}))
        self.assertTrue(result.evaluate({"param1":0, "param2":1}))
        self.assertFalse(result.evaluate({"param1":0, "param2":0}))
    
    def testXor(self):
        result = self.t1 ^ self.t2
        self.assertIsInstance(result, Expression)
        self.assertFalse(result.evaluate({"param1":1, "param2":1}))
        self.assertTrue(result.evaluate({"param1":1, "param2":0}))
        self.assertTrue(result.evaluate({"param1":0, "param2":1}))
        self.assertFalse(result.evaluate({"param1":0, "param2":0}))
    
    def testNot(self):
        result = -(self.t1 & self.t2)
        self.assertIsInstance(result, Expression)
        self.assertFalse(result.evaluate({"param1":1, "param2":1}))
        self.assertTrue(result.evaluate({"param1":1, "param2":0}))
        self.assertTrue(result.evaluate({"param1":0, "param2":1}))
        self.assertTrue(result.evaluate({"param1":0, "param2":0}))

    def testNesting(self):
        result = (self.t1 & self.t2) | self.t3
        self.assertTrue(result.evaluate({"param1":1, "param2":1, "param3":1}))
        self.assertTrue(result.evaluate({"param1":1, "param2":1, "param3":0}))
        thing = {"param1":1, "param2":0, "param3":1}
        self.assertTrue(result.evaluate({"param1":1, "param2":0, "param3":1}))
        self.assertFalse(result.evaluate({"param1":1, "param2":0, "param3":0}))
        self.assertTrue(result.evaluate({"param1":0, "param2":1, "param3":1}))
        self.assertFalse(result.evaluate({"param1":0, "param2":1, "param3":0}))
        self.assertTrue(result.evaluate({"param1":0, "param2":0, "param3":1}))
        self.assertFalse(result.evaluate({"param1":0, "param2":0, "param3":0}))

class TermTest(unittest.TestCase):
    """ Test that each term in the expression itself is correct."""

    def setUp(self):
        self.uut = Term("param1", 1)

    def testDefault(self):
        val = self.uut.evaluate({}, default=True)
        self.assertTrue(self.uut.evaluate({}, default=True))
        self.assertFalse(self.uut.evaluate({}, default=False))

    def testEvaluate(self):
        self.assertTrue(self.uut.evaluate({"param1": 1}))
        self.assertFalse(self.uut.evaluate({"param1": 0}))

    def testInvert(self):
        result = -self.uut
        self.assertFalse(result.evaluate({"param1": 1}))
        self.assertTrue(result.evaluate({"param1": 0}))

class AndListTest(unittest.TestCase):
    """ Tests for the AndList class."""

    def setUp(self):
        self.uut = AndList({"param1":1, "param2":1, "param3":1})

    def testAndList(self):
        self.assertTrue(self.uut.evaluate({"param1":1, "param2":1, "param3":1}))
        self.assertFalse(self.uut.evaluate({"param1":1, "param2":1, "param3":0}))
        self.assertFalse(self.uut.evaluate({"param1":1, "param2":0, "param3":1}))
        self.assertFalse(self.uut.evaluate({"param1":1, "param2":0, "param3":0}))
        self.assertFalse(self.uut.evaluate({"param1":0, "param2":1, "param3":1}))
        self.assertFalse(self.uut.evaluate({"param1":0, "param2":1, "param3":0}))
        self.assertFalse(self.uut.evaluate({"param1":0, "param2":0, "param3":1}))
        self.assertFalse(self.uut.evaluate({"param1":0, "param2":0, "param3":0}))

class OrListTest(unittest.TestCase):
    """ Tests for the OrList class."""

    def setUp(self):
        self.uut = OrList({"param1":1, "param2":1, "param3":1})

    def testAndList(self):
        self.assertTrue(self.uut.evaluate({"param1":1, "param2":1, "param3":1}))
        self.assertTrue(self.uut.evaluate({"param1":1, "param2":1, "param3":0}))
        self.assertTrue(self.uut.evaluate({"param1":1, "param2":0, "param3":1}))
        self.assertTrue(self.uut.evaluate({"param1":1, "param2":0, "param3":0}))
        self.assertTrue(self.uut.evaluate({"param1":0, "param2":1, "param3":1}))
        self.assertTrue(self.uut.evaluate({"param1":0, "param2":1, "param3":0}))
        self.assertTrue(self.uut.evaluate({"param1":0, "param2":0, "param3":1}))
        self.assertFalse(self.uut.evaluate({"param1":0, "param2":0, "param3":0}))
