from twisted.python import util
from twisted.internet.defer import Deferred

from zope.interface import Interface, implements

from nevow import rend, static, url
from nevow import tags as t
from nevow.stan import directive
from nevow.loaders import xmlfile, stan

from safe.webmanager.template import *
from safe.webmanager.results import ResultsPage
from safe.models import DesignPoint

class DesignPointPage(TemplatedPage):
    implements(ITemplatedPage)

    page_title = "Design Point"

    class Content(rend.Fragment):
        """Fragment to generate the content of the page."""

        docFactory = stan([
            t.invisible(data=directive("design_point"),
                render=directive("design_point"))[
                    t.p[
                        t.b["Design Point id: "],
                        t.slot(name="id")
                    ],
                    t.p[
                        t.b["Completed: "],
                        t.slot(name="completed")
                    ],
                    t.p[
                        t.slot(name="results")
                    ],
                    t.pre[t.slot(name="xml")],
            ]
        ])

        def __init__(self, experiment_id, design_point_id, dbpool):
            """Create the fragment which displays a design_point."""

            self.experiment_id = experiment_id
            self.design_point_id = design_point_id
            self.dbpool = dbpool

        def render_design_point(self, ctx, data):
            """Fill the design_point slots."""

            ctx.fillSlots('id',data[0])
            ctx.fillSlots('completed',data[1])
            ctx.fillSlots('results',
                    t.a(href=url.here.child('results'))["Results"])
            ctx.fillSlots('xml',data[2])

            return ctx

        def data_design_point(self, ctx, data):
            """Query to find the specified design_point."""

            def sanitize(results):
                """Sanitize the data from the database for display.

                For now, just callback the result as just the first row in the
                result set.
                
                """

                if len(results) != 1:
                    answer.errback("Design Point not found")
                else:
                    answer.callback(results[0])

            query = """
            SELECT id, completed, config
            FROM design_points 
            WHERE id = %d
            """ % (self.design_point_id)

            answer = Deferred()
            self.dbpool.runQuery(query).addCallback(sanitize)

            return answer

    def render_content(self, ctx, data):
        """Render the content of the page."""

        return ctx.tag[self.Content(self.experiment_id,
            self.design_point_id,self.dbpool)]

    def __init__(self, experiment, design_point, dbpool):
        """Constructor for the DesignPoint() page."""

        super(TemplatedPage, self).__init__()
        self.experiment = experiment
        self.design_point = design_point
        self.design_point_id = design_point.cachedValue("id")
        self.experiment_id = experiment.cachedValue("id")
        self.dbpool = dbpool
        self.children = {'results' :
            ResultsPage(self.dbpool,
                experiment = self.experiment,
                design_point = self.design_point)
        }

class DesignPointsPage(TemplatedPage):
    implements(ITemplatedPage)

    page_title = "Design Points"
    
    class Content(rend.Fragment):

        docFactory = stan(
            t.table(data=directive("experiments"), render=directive("sequence"))[
                t.tr(pattern="header")[
                    t.th["Id"],
                    t.th["XML"],
                    t.th["Completed"],
                ],
                t.tr(pattern="item", render=directive("row"))[
                    t.td[t.slot(name="id")],
                    t.td[t.pre[t.slot(name="xml")]],
                    t.td[t.slot(name='completed')],
                ]
            ]
        )
        
        def __init__(self, experiment, dbpool):
            """Initalize the Experiment Content fragment with the db."""

            self.dbpool = dbpool
            self.experiment = experiment
            self.experiment_id = experiment.cachedValue("id")

        def render_row(self, context, data):
            """Create each row of the table of experiments."""
            dp_path = '/experiments/%d/design_points/%d/' %\
                (self.experiment_id, data[0])
            context.fillSlots('id',t.a(href=dp_path)[data[0]])
            context.fillSlots('xml',data[1])
            context.fillSlots('completed',data[2])
            return context

        def data_experiments(self,context,data):
            """Method to get the experiments in the database"""

            #take out the hardcoded experiment
            query = """
            SELECT id,config,completed
            FROM design_points
            WHERE experiment_id = %d
            """ % (self.experiment_id)

            return self.dbpool.runQuery(query)

    def __init__(self, experiment, dbpool):
        super(TemplatedPage, self).__init__()
        self.dbpool = dbpool
        self.experiment = experiment
        self.experiment_id = experiment.cachedValue("id")
    
    def render_content(self,ctx,data):
        """Return the content to be placed in the global template."""
        
        return ctx.tag[self.Content(self.experiment, self.dbpool)]

    def childFactory(self, ctx, segment):
        if segment.isdigit():
            d = DesignPoint.find(self.dbpool, id = int(segment))
            d.addCallback(lambda s: DesignPointPage(
                self.experiment, s, self.dbpool))
            return d
