# -*- test-case-name: safe.test.test_messages -*-

""" messages.py: A module with classes and methods for encoding and decoding
    messages between the simulation clients and the execution manager.
"""

__author__  = "Bryan C. Ward"
__email__   = "bryan.ward@bucknell.edu"

from struct import pack,unpack,calcsize
from operator import itemgetter as _itemgetter

REGISTER = 0
NEXT_REQUEST = 1
NEXT_REPLY = 2
SAMPLE = 3 
FINISHED = 4 
TERMINATE = 5
HASH = 6

def decode(data):
    """
    Decode a stream of data of arbitrary length into the different message
    types. If the data contains a valid message at the head, pop it off and
    return the message class for consumption while also returning the rest of
    the data stream. Otherwise return the data unchanged with None for the
    message. For example::

        msg,data = decode(data)

    ``msg`` will be the parsed message and ``data`` will have the message popped off.
    """
    if len(data) < calcsize('!I'):
        return None,data
    else:
        (opcode,) = unpack('!I',data[:4])

    if opcode == REGISTER and len(data) >= Register.size():
        return Register.decode(data[:Register.size()]),data[Register.size():]
    elif opcode == NEXT_REQUEST and len(data) >= NextRequest.size():
        return (NextRequest.decode(data[:NextRequest.size()]),
                data[NextRequest.size():])
    elif opcode == NEXT_REPLY and len(data) > calcsize(NextReply._header):
      #NEXT_REPLY is handled differently because it has a variable length
      msg = NextReply.decode(data)
        if msg == None:
            return None,data
        else:
            return msg,data[len(msg):]
    elif opcode == SAMPLE and len(data) >= Sample.size():
        return Sample.decode(data[:Sample.size()]),data[Sample.size():] 
    elif opcode == HASH and len(data) >= Hash.size():
        return Hash.decode(data[:Hash.size()]),data[Hash.size():]
    elif opcode == FINISHED and len(data) >= Finished.size():
        return Finished.decode(data[:Finished.size()]),data[Finished.size():]
    elif opcode == TERMINATE and len(data) >= Terminate.size():
        return Terminate.decode(data[:Terminate.size()]),data[Terminate.size():]
    elif opcode not in [REGISTER,NEXT_REQUEST,NEXT_REPLY,SAMPLE,HASH,FINISHED,TERMINATE]:
        #invalid opcode
        raise MessageDecodeError("Invalid Opcode %d" % opcode)
    else:
        #valid opcode, but not enough data
        # return None for our message and the return the data unchanged
        return None,data

class Register(str):
    """Simple class for the representation of the ``REGISTER`` message."""

    _format = '!I128s'

    __slots__ = ()

    def __new__(_cls,info):
        return str.__new__(_cls,info)

    def encode(self):
        """Encode the data to transmit on the network."""

        return pack(self._format,REGISTER,self)

    def __repr__(self):
        return "REGISTER: %s" % self

    def __len__(self):
        return calcsize(self._format)

    @classmethod
    def decode(_cls,data,new=str.__new__):
        """Decode a ``REGISTER`` message from the network.

        If the data is not a valid message, raise an exception.

        """
        (opcode,info) = unpack(Register._format,data)
        if opcode == REGISTER:
            return new(_cls,info)
        else:
            raise MessageDecodeError("Expected opcode %d found %d"\
                    % (REGISTER,opcode))

    @classmethod
    def size(_cls):
        """Return the size of the encoded message string."""
        return calcsize(Register._format)

    info = property(_itemgetter(slice(0,None)))

class NextRequest(tuple):
    """Message for requesting a new simulation."""

    _format = '!I'

    __slots__ = ()

    def __new__(_cls):
        return tuple.__new__(_cls,())

    def encode(self):
        """Encode the data to transmit on the network."""

        return pack(self._format,NEXT_REQUEST)

    def __repr__(self):
        return "NEXT_REQUEST"

    def __len__(self):
        """Return the length of the encoded message."""

        return calcsize(self._format)

    @classmethod
    def decode(_cls,data,new=tuple.__new__):
        """Decode a ``NEXT_REQUEST`` message from the network.

        If the data is not a valid message, raise an exception.

        """
        
        (opcode,) = unpack(NextRequest._format,data)
        if opcode == NEXT_REQUEST:
            return new(_cls,())
        else:
            raise MessageDecodeError("Expected opcode %d found %d"\
                    % (NEXT_REQUEST,opcode))

    @classmethod
    def size(_cls):
        """Return the size of the encoded message string."""
        return calcsize(NextRequest._format)

class NextReply(str):
    """Message for the response of a request of a simulation."""

    _header = "!II"

    _format = "!II%ds"

    __slots__ = ()

    def __new__(_cls,xml):
        return str.__new__(_cls,xml)

    def __repr__(self):
        return "NEXT_REPLY: %s" % self

    def __len__(self):
        """Return the length of the encoded message."""

        return calcsize("!II%ds" % (len(self.xml)))

    def encode(self):
        """Encode the data to transmit on the network."""

        return pack(NextReply._format % (len(self.xml)),NEXT_REPLY,len(self.xml),self)

    @classmethod
    def decode(_cls,data,new=str.__new__):
        """Decode a ``NEXT_REPLY`` message from the network.

        If the data is not a valid message, raise an exception.

        """

        (opcode,length) = unpack(NextReply._header,
                data[:calcsize(NextReply._header)])
        if opcode == NEXT_REPLY:
            if len(data) < calcsize("!II%ds" % (length)):
                return None
            else:
                (opcode,xml_len,xml) = unpack(NextReply._format % (length),
                            data[:calcsize(NextReply._header)+length])
                return new(_cls,xml)
        else:
            raise MessageDecodeError("Expected opcode %d found %d"\
                    % (NEXT_REPLY,opcode))

    xml = property(_itemgetter(slice(0,None)))

class Sample(tuple):
    """Message containing a statistical sample for transmission."""

    _format = '!Id64s8sd'

    __slots__ = ()

    def __new__(_cls,timestamp,object,metric,value):
        return tuple.__new__(_cls,(timestamp,object,metric,value))

    def encode(self):
        """Encode the data to transmit on the network."""

        return pack(self._format,SAMPLE,self[0],self[1],self[2],self[3])

    def __repr__(self):
        return "SAMPLE: <%f, %s, %s, %f>" % self

    def __len__(self):
        """Return the length of the encoded message."""

        return calcsize(self._format)

    @classmethod
    def decode(_cls,data,new=tuple.__new__):
        """Decode a ``SAMPLE`` message from the network.

        If the data is not a valid message, raise an exception.

        """

        (opcode,timestamp,object,metric,value) = unpack(Sample._format,data)
        if opcode == SAMPLE:
            return new(_cls,(timestamp,object,metric,value))
        else:
            raise MessageDecodeError("Expected opcode %d found %d"\
                    % (SAMPLE,opcode))

    @classmethod
    def size(_cls):
        """Return the length of an encoded ``SAMPLE`` message."""

        return calcsize(Sample._format)


    timestamp = property(_itemgetter(0))
    object = property(_itemgetter(1))
    metric = property(_itemgetter(2))
    value = property(_itemgetter(3))

class Finished(tuple):
    """Message indicating a simulation is completed."""

    _format = '!I'

    __slots__ = ()

    def __new__(_cls):
        return tuple.__new__(_cls,())

    def encode(self):
        """Encode the data to transmit on the network."""

        return pack(self._format,FINISHED)

    def __repr__(self):
        return "FINISHED"

    def __len__(self):
        """Return the length of an encoded ``FINISHED`` message."""

        return calcsize(self._format)

    @classmethod
    def decode(_cls,data,new=tuple.__new__):
        """Decode a ``FINISHED`` message.

        If the data is not a valid message, raise an exception.

        """

        (opcode,) = unpack(Finished._format,data)
        if opcode == FINISHED:
            return new(_cls,())
        else:
            raise MessageDecodeError("Expected opcode %d found %d"\
                    %(FINISHED,opcode))

    @classmethod
    def size(_cls):
        """Return the length of an encoded ``FINISHED`` message."""
        return calcsize(Finished._format)

class Terminate(tuple):
    """Message indicating an experiment is completed."""

    _format = '!I'

    __slots__ = ()

    def __new__(_cls):
        return tuple.__new__(_cls,())

    def encode(self):
        """Encode the message to transmit on the network."""

        return pack(self._format,TERMINATE)

    def __repr__(self):
        return "TERMINATE"

    def __len__(self):
        """Return the length of the encoded message."""
        return calcsize(self._format)
    
    @classmethod
    def decode(_cls,data,new=tuple.__new__):
        """Decode a ``TERMINATE`` message.

        If the data is not a valid message, raise an exception.

        """

        (opcode,) = unpack(Terminate._format,data)
        if opcode == TERMINATE:
            return new(_cls,())
        else:
            raise MessageDecodeError("Expected opcode %d found %d"\
                    %(TERMINATE,opcode))

    @classmethod
    def size(_cls):
        """Return the length of an encoded ``TERMINATE`` message."""

        return calcsize(Terminate._format)


class Hash(str):
    """Message containing a hash for transmission."""

    _format = '!I64s'

    __slots__ = ()

    def __new__(_cls,hash):
        return str.__new__(_cls,hash)

    def encode(self):
        """Encode the data to transmit on the network."""

        return pack(self._format,HASH,self)

    def __repr__(self):
        return "HASH: <%s>" % self

    def __len__(self):
        """Return the length of the encoded message."""

        return calcsize(self._format)

    @classmethod
    def decode(_cls,data,new=str.__new__):
        """Decode a ``HASH`` message from the network.

        If the data is not a valid message, raise an exception.

        """

        (opcode,hash) = unpack(Hash._format,data)
        if opcode == HASH:
            return new(_cls,(hash))
        else:
            raise MessageDecodeError("Expected opcode %d found %d"\
                    % (HASH,opcode))

    @classmethod
    def size(_cls):
        """Return the length of an encoded ``HASH`` message."""

        return calcsize(Hash._format)


    hash = property(_itemgetter(slice(0,None)))
