#!/usr/bin/env python
# import argparse
from optparse import OptionParser
from safe.boolean import *
from safe.exceptions import UndeclaredFactorException
import experiment
import logging
import handlers
import sys
import os

class ColorFormatter(logging.Formatter):

    def format(self, record):
        if record.levelname == 'WARNING':
            begin_color = bcolors.WARNING
            end_color = bcolors.ENDC
        elif record.levelname == 'INFO':
            begin_color = bcolors.ENDC
            end_color = bcolors.ENDC
        elif record.levelname == 'ERROR':
            begin_color = bcolors.ERROR
            end_color = bcolors.ENDC
        elif record.levelname == 'CRITICAL':
            begin_color = bcolors.CRITICAL
            end_color = bcolors.ENDC
        else:
            begin_color = ''
            end_color = ''

        self._fmt = begin_color + '%(levelname)8s : %(msg)s' + end_color
        s = logging.Formatter.format(self, record)
        return s


class bcolors:
    # ANSI color codes:
    # 0 : Black
    # 1 : Red 
    # 2 : Green
    # 3 : Yellow
    # 4 : Blue
    # 5 : Magenta
    # 6 : Cyan
    # 7 : White
    # 
    # \033[31m  : foreground red
    # \033[91m  : foreground bright red
    # \033[41m  : background red
    # \033[101m : background bright red
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'

    WARNING = '\033[30m\033[107m'
    ERROR = '\033[97m\033[44m'
    ERROR = '\033[97m\033[101m'

    ENDC = '\033[0m'

    def disable(self):
        self.HEADER = ''
        self.OKBLUE = ''
        self.OKGREEN = ''
        self.WARNING = ''
        self.FAIL = ''
        self.ENDC = ''

class ExpValidator(object):

    def __init__(self, xml):
        self.xml = xml
        self.factor_list = []
        self.exp = experiment.CreateFromDocument(self.xml)
        self.mapped_values = {}
        self.restriction_expression = None
        self.allowed_tests = ['EQUALS', 'LT', 'GT']
        self.allowed_operators = \
        ['MULT', 'FDIV', 'IDIV', 'PLUS', 'MINUS', 'MOD', 'POW']

    def validate(self):
        for factor in self.exp.factorlist.factor:
            logger.info('Found factor \'%s\'' % str(factor))
            self.factor_list.append(str(factor))

        self.validate_conditions()

    def validate_conditions(self):
        self.validate_memberof()
        self.validate_sequence()
        self.validate_linking_restriction()
        self.validate_exclusion_restriction()


    def validate_memberof(self):
        for memberof in self.exp.conditions.memberof:
            if str(memberof.factor) in self.factor_list:
                # Assert: A valid factor name was provided.
                logger.info('Validating <memberof> element where...')
                logger.info('\t<factor> = %s' % str(memberof.factor))
                logger.info('\t<listid> = %s' % str(memberof.listid))

                found_internal_list = False
                found_external_list = False
                # Iterate through all of the list elements in the experiment
                # space.  If one of them has an 'id' attribute which matches
                # our list_identifier, use that.  Otherwise, assume the
                # experimenter was specifying an external file.
                for levellist in self.exp.conditions.levellist:
                    if levellist.id == memberof.listid:
                        values = map(str, levellist.level)
                        found_internal_list = True
                        # We've found the right list.  Break out of loop.
                        break

                # Check if there is an external file with the listid as
                # filename.  Assume they had the .xml extension already on
                # there
                if os.path.exists(str(memberof.listid)):
                    found_external_list = True

                logger.info('\t<listid> matches an internal <levellist> \'id\' attribute: %s' % found_internal_list)
                logger.info('\t<listid> matches external filename: %s' % found_external_list)

                if found_internal_list and found_external_list:
                    logger.warn("""list identifier \'%s\' matches both an
                    internal <levellist> element and an external filename.
                    The parser will use the internally-defined list.""" %
                    memberof.listid)
                elif not(found_internal_list) and not(found_external_list):
                    logger.error("""list identifier \'%s\' does not match
                    the \'id\' attribute of any any internal <levellist>
                    element, nor does it match an external filename in
                    this directory.""" % memberof.listid)

            else:
                # Assert: An invalid factor name was provided.
                logger.error("""Factor \'%s\' in <memberof> element does not
                appear in <factorlist>.""" % str(memberof.factor))

    def ic_message(self, content, element):
        """A method for resturning an "Invalid Content" message.  Returns
        a string of the form
        '%s is not valid content for the %s element' % (content, element)
        """
        msg = '\'%s\' is not valid content for the %s element'
        msg = msg % (str(content), str(element))
        return msg

    def validate_expression(self, expression, var_name, pad = ''):
        # Check if <lconst> exists and if it's valid.
        if expression.lconst != None:
            logger.info('%s<lconst> = %s' % (pad, str(expression.lconst)))
            try:
                float(expression.lconst)
            except ValueError:
                content = expression.lconst
                element = '<lconst>'
                msg = self.ic_message(content, element)
                logger.error(msg)

        # Check if <lvar> exists and if it's valid.
        if expression.lvar != None:
            logger.info('%s<lvar> = %s' % (pad, str(expression.lvar)))
            try:
                float(expression.lvar)
                # If we got to this point they're trying to use a
                # number as a var.
                content = expression.lvar
                element = '<lvar>'
                msg = self.ic_message(content, element)
                logger.error(msg)
            except ValueError:
                # Assert: they're using a valid string for the var name.
                if var_name != None and var_name != str(expression.lvar):
                    # Assert: a variable name was previously used but does not
                    # match this name.
                    logger.warn("""Variable name \'%s\' used in <lvar>
                    does not match the variable name \'%s\' used
                    previously.  The parser uses a single variable for an
                    entire expression, regardless of the value of any
                    <lvar> and <rvar> elements.""" %
                    (str(expression.lvar), var_name))
                if var_name == None:
                    var_name = str(expression.lvar)

        # Check if <lexpr> exists and if it's valid.
        if expression.lexpr != None:
            logger.info('%sValidating <lexpr> where...' % pad)
            self.validate_expression(expression.lexpr, var_name, pad + '    ')

        # Check if <op> is valid.
        logger.info('%s<op> = %s' % (pad, str(expression.op)))
        if not(str(expression.op) in self.allowed_operators):
            content = expression.op
            element = '<op>'
            msg = self.ic_message(content, element)
            logger.error(msg)

        # Check if <rconst> exists and if it's valid.
        if expression.rconst != None:
            logger.info('%s<rconst> = %s' % (pad, str(expression.rconst)))
            try:
                float(expression.rconst)
            except ValueError:
                content = expression.rconst
                element = '<rconst>'
                msg = self.ic_message(content, element)
                logger.error(msg)

        # Check if <rvar> exists and if it's valid.
        if expression.rvar != None:
            logger.info('%s<rvar> = %s' % (pad, str(expression.rvar)))
            try:
                float(expression.rvar)
                # If we got to this point they're trying to use a
                # number as a var.
                content = expression.rvar
                element = '<rvar>'
                msg = self.ic_message(content, element)
                logger.error(msg)
            except ValueError:
                # Assert: they're using a valid string for the var name.
                if var_name != None and var_name != str(expression.rvar):
                    # Assert: a variable name was previously used but does not
                    # match this name.
                    logger.warn("""Variable name \'%s\' used in <rvar>
                    does not match the variable name \'%s\' used
                    previously.  The parser uses a single variable for an
                    entire expression, regardless of the value of any
                    <lvar> and <rvar> elements.""" %
                    (str(expression.lvar), var_name))
                if var_name == None:
                    var_name = str(expression.rvar)

        # Check if <rexpr> exists and if it's valid.
        if expression.rexpr != None:
            logger.info('%sValidating <rexpr> where...' % pad)
            self.validate_expression(expression.rexpr, var_name, pad + '    ')

        # Check if there is more than one left- or right-hand piece to
        # the expression.
        num_lefts  = expression.lvar != None
        num_lefts += expression.lconst != None
        num_lefts += expression.lexpr != None

        num_rights  = expression.rvar != None
        num_rights += expression.rconst != None
        num_rights += expression.rexpr != None

        if num_lefts < 1:
            logger.error("""A left-hand side to the expression is
            missing.""")

        if num_lefts > 1:
            logger.error("""There are too many left-hand pieces to
            the expression""")

        if num_rights < 1:
            logger.error("""A right-hand side to the expression is
            missing.""")

        if num_rights > 1:
            logger.error("""There are too many right-hand pieces to
            the expression""")

    def validate_sequence(self, pad = ''):
        for sequence in self.exp.conditions.sequence: 
            logger.info('Validating <sequence> element where...')

            # Check if <factor> is valid.
            logger.info('%s<factor> = %s' % (pad, str(sequence.factor)))
            if not(str(sequence.factor) in self.factor_list):
                logger.error("""Sequence factor \'%s\' does not
                appear in <factorlist> and is not valid.  Must be one
                of %s""" % (sequence.factor, self.factor_list))

            # Check if <test> is valid.
            logger.info('%s<test> = %s' % (pad, str(sequence.test)))
            if not(str(sequence.test) in self.allowed_tests):
                logger.error("""Sequence test \'%s\' is not a valid
                test.  Must one of %s""" % (sequence.test,
                self.allowed_tests))

            # Check if <lconst> exists and if it's valid.
            if sequence.lconst != None:
                logger.info('%s<lconst> = %s' % (pad, str(sequence.lconst)))
                try:
                    float(sequence.lconst)
                    # Note: values in scientific notation are allowed with
                    # this (e.g., 8.123e4)
                except ValueError:
                    content = sequence.lconst
                    element = '<lconst>'
                    msg = self.ic_message(content, element)
                    logger.error(msg)

            # Check if <lvar> exists and if it's valid.
            var_name = None
            if sequence.lvar != None:
                logger.info('%s<lvar> = %s' % (pad, str(sequence.lvar)))
                try:
                    float(sequence.lvar)
                    # If we got to this point they're trying to use a
                    # number as a var.
                    content = sequence.lvar
                    element = '<lvar>'
                    msg = self.ic_message(content, element)
                    logger.error(msg)
                except ValueError:
                    # If we got to this point they've used some kind of
                    # string as the var.
                    var_name = str(sequence.lvar)

            # Check if <lexpr> exists and if it's valid.
            if sequence.lexpr != None:
                logger.info('%sValidating <lexpr> where...' % pad)
                self.validate_expression(sequence.lexpr, var_name, pad + '    ')

            # Check if <op> is valid.
            logger.info('%s<op> = %s' % (pad, str(sequence.op)))
            if not(str(sequence.op) in self.allowed_operators):
                content = sequence.op
                element = '<op>'
                msg = self.ic_message(content, element)
                logger.error(msg)

            # Check if <rconst> exists and if it's valid.
            if sequence.rconst != None:
                logger.info('%s<rconst> = %s' % (pad, str(sequence.rconst)))
                try:
                    float(sequence.rconst)
                except ValueError:
                    content = sequence.rconst
                    element = '<rconst>'
                    msg = self.ic_message(content, element)
                    logger.error(msg)

            # Check if <rvar> exists and if it's valid.
            if sequence.rvar != None:
                logger.info('%s<rvar> = %s' % (pad, str(sequence.rvar)))
                try:
                    float(sequence.rvar)
                    # If we got to this point they're trying to use a
                    # number as a var.
                    content = sequence.rvar
                    element = '<rvar>'
                    msg = self.ic_message(content, element)
                    logger.error(msg)
                except ValueError:
                    # Assert: they're using a valid string a var name.
                    if var_name != None and var_name != str(sequence.rvar):
                        # Assert: <lvar> was used previously but the variable
                        # named used in <lvar> does not match this variable
                        # name in <rvar>
                        logger.warn("""Variable name \'%s\' used in <rvar>
                        does not match the <lvar> \'%s\' used previously.
                        The parser uses a single variable for an entire
                        expression, regardless of the value of the <lvar>
                        and <rvar> elements.""" %
                        (str(sequence.rvar), var_name))

            # Check if <rexpr> exists and if it's valid.
            if sequence.rexpr != None:
                logger.info('%sValidating <rexpr> where...' % pad)
                self.validate_expression(sequence.rexpr, var_name, pad + '    ')

            # Check if there is more than one left- or right-hand piece to
            # the expression.
            num_lefts  = sequence.lvar != None
            num_lefts += sequence.lconst != None
            num_lefts += sequence.lexpr != None

            num_rights  = sequence.rvar != None
            num_rights += sequence.rconst != None
            num_rights += sequence.rexpr != None

            if num_lefts < 1:
                logger.error("""A left-hand side to the expression is
                missing.""")

            if num_lefts > 1:
                logger.error("""There are too many left-hand pieces to
                the expression""")

            if num_rights < 1:
                logger.error("""A right-hand side to the expression is
                missing.""")

            if num_rights > 1:
                logger.error("""There are too many right-hand pieces to
                the expression""")

    def validate_linking_restriction(self):
        
        for lr in self.exp.conditions.linkingrestriction:
            logger.info('Validating <linkingrestriction> element where...')
            # Check if each <factor> is valid.
            for factor in lr.factor:
                logger.info('\t<factor> = %s' % str(factor))
                if not(str(factor) in self.factor_list):
                    logger.error("""Linking restriction factor \'%s\' does
                    not appear in <factorlist> and is not valid.  Must be
                    one of %s""" % (factor, self.factor_list))
            # TODO: figure out a way to check the list sizes, or output a
            # better message.
            logger.info("""Note that in the case of differing level list sizes among
            linked factors, the smallest level list determines the number of
            linked combinations.""")

    def validate_exclusion_restriction(self):
        for er in self.exp.conditions.exclusionrestriction:
            logger.info('Validating <exclusionrestriction> element where...')
            for setting in er.setting:
                logger.info('\t<factor> = %s' % str(setting.factor))
                logger.info('\t<level> = %s' % str(setting.level))
                if not(str(setting.factor) in self.factor_list):
                    logger.error("""Exclusion restriction factor \'%s\' does
                    not appear in <factorlist> and is not valid.  Must be
                    one of %s""" % (setting.factor, self.factor_list))

        

if __name__ == '__main__':


    usage = './%prog EXPERIMENT_DESCRIPTION.xml [options]'
    parser = OptionParser(usage = usage)
    parser.add_option('-v', '--verbose', dest='verbose',
                        action='store_true', default=False,
                        help='print verbose output')
    (options, args) = parser.parse_args()

    # Creating logger.
    logger = logging.getLogger('expvalidator_logger')
    logger.setLevel(logging.INFO)

    formatter = logging.Formatter('\033[93m%(levelname)8s : %(msg)s')
    cformatter = ColorFormatter()

    ch = logging.StreamHandler()
    ch.setFormatter(cformatter)
    if (options.verbose):
        ch.setLevel(logging.INFO)
        logger.addHandler(ch)
    else:
        ch.setLevel(logging.ERROR)
        logger.addHandler(ch)


    xml_string = file(args[0]).read()
    v = ExpValidator(xml_string)
    v.validate()
