import stats
import os
import json
from twisted.internet.protocol import Protocol, Factory
from twisted.protocols.basic import Int32StringReceiver

from django.conf import settings
from django.db.models import Count, StdDev, Avg
os.environ.setdefault("DJANGO_SETTINGS_MODULE","web.config.settings")
from web.safe.models import *

from fabric.colors import blue, yellow


class AnalysisManager(Int32StringReceiver):
    """Manager for a Global Analyzer Client

    The class extends the :class:`Int32StringReceiver` class and overrides the
    :func:`dataReceived` method. This allows it to accept messages from the
    remote client in an asynchronous fashion. A single instance of this class
    manages a single client connection, so any state variables specific to this
    instance are for a given client. The ``self.factory`` variable is the
    object which holds any information for the aggregate of all
    :class:`AnalysisManager` instances."""

    def __init__(self, factory):
        self.factory=factory
        self.handlers = {"estimate" : self._handleEstimate}

    def stringReceived(self,raw_data):
        """Method invoked when data is received on the network."""
        #Decode json data to a dictionary
        data = json.loads(raw_data)

        #Call the correct handler function
        if data['type'] in self.handlers.keys():
            self.handlers[data['type']](data)
        else:
            raise Exception("Invalid message received: "+str(data))


    def connectionMade(self):
        """Method invoked when a connection to the client is made."""
        print("New connected client ")

    def connectionLost(self,reason):
        """Method invoked when the connection to the client is lost.

        When the connection is lost, ``self`` should be removed from the list
        of active protocols. In the case that all protocols have terminated
        their connections and there are no design points left to simulate, the
        execution manager can terminate.

        """
        #sys.stderr.write("Connection Lost: %s\n" % self.transport.getPeer())
        #sys.stderr.write(str(reason)+"\n")

           
    def _handleEstimate(self,msg):
        """Method invoked when an estimate string is received"""
        if self.factory.dp is None or int(self.factory.dp.id) != int(msg['dp_id']) :
            return
        metric_id = str(msg['metric'])
        self.factory.estimates[metric_id] = self.factory.estimates.get(metric_id, [])
        self.factory.estimates[metric_id].append((float(msg["mean"]), int(msg["results"])))
        self.factory.count[metric_id] = self.factory.count.get(metric_id, 0) + 1
        #If we have enough estimates, right now 4 for each metric
        if len([ 1 for c in self.factory.count.values() if c > 4 ]) == len(self.factory.count):
            self.factory.count = {}
            if self._analyze():
                self.factory.estimates = {}
                self.factory.dp.finished = 1
                self.factory.dp.save()
                self.factory.dp = self.factory.experiment.getCurrentDesignPoint()
                
    def _analyze(self):
        """Analyzes the current set of estimates, and returns true if the relative precision
        of each metric (calculated using confidence intervals) is less than the specified precision"""
        #using akaroa2-based algorthm
        for metric in self.factory.estimates.values():
            mean = 0
            n = 0
            variance = 0
            for x_j, n_j in metric:
                mean += n_j * x_j
                n += n_j
                variance += x_j * n_j**2

            mean /= float(n)
            variance /= float(n**2)
            stddev = variance ** 0.5
            if mean == 0:
                print("Mean is zero, continuing")
                return None
            m = len(metric)
            t_star = stats.student_idf(m - 1, self.factory.conf_level * 0.5  + 0.5)
            ci_size = (t_star * stddev) / (m**0.5)
            print(ci_size / mean)
            if (ci_size / mean) > self.factory.precision:
                return False
        return True


class AnalysisFactory(Factory):
  """Class which generates specific protocols objects for each connection.

  The :class:`AnalysisFactory` is responsible for creating all of the
  :class:`AnalysisManager` instances. This is done in the
  :func:`buildProtocol` method which has been overridden from the
  :class:`Factory` superclass.
  """
  def __init__(self, experiment, conf_level, precision, min_threshold):
    """Initialization of the :class:`AnalysisFactory`.

    This method receives a set of parameters and initializes several variables that will be used by the globalAnalyzer to 
    compute the relative precision and determine when a simulation should end."""
    self.experiment = experiment
    self.conf_level = conf_level
    self.precision = precision
    self.min_threshold = min_threshold
    self.estimates = {}
    self.count = {}
    self.dp = self.experiment.getCurrentDesignPoint()

  def buildProtocol(self,addr):
    """Build the :class:`AnalysisManager` protocol.

    This method overrides the default :func:`buildProtocol` method and must
    return a protocol subclass."""
    print("Building a protocol")
    return AnalysisManager(self)
