from django.db import models
from django import forms
from django.contrib.auth.forms import UserCreationForm, AuthenticationForm
from django.contrib.auth.models import User
from django.contrib.auth import authenticate
from safe.models import Experiment, ExpDescription, DesignPoint, SimulationRun
import re

from crispy_forms.helper import FormHelper

"""
Forms for input within the user interface 
"""

# Form with account information
class AccountForm(forms.Form):
    password = forms.CharField(widget=forms.PasswordInput)
    new_password = forms.CharField(widget=forms.PasswordInput, required=False)
    verify_password = forms.CharField(widget=forms.PasswordInput, required=False)

    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop("user", None)
        super(AccountForm, self).__init__(*args, **kwargs)

    def clean_password(self):
        password = self.cleaned_data["password"]
        if self.user == None:
            raise forms.ValidationError("Invalid Password")
        
        auth = authenticate(username=self.user.username, password=password)
        if auth == None:
            raise forms.ValidationError("Invalid Password")
        return password

    def clean_verify_password(self):
        pw1 = self.cleaned_data["new_password"]
        pw2 = self.cleaned_data["verify_password"]
        if pw1 != pw2:
            raise forms.ValidationError("Passwords Must Match")
        return pw1

# User creation form
class UserCreateForm(forms.ModelForm):
    password = forms.CharField(widget=forms.PasswordInput)
    verify_password = forms.CharField(widget=forms.PasswordInput)

    def clean_verify_password(self):
        pw1 = self.cleaned_data["password"]
        pw2 = self.cleaned_data["verify_password"]
        if pw1 != pw2:
            raise forms.ValidationError("Passwords Must Match")
        return pw1
            
    class Meta:
        model = User
        fields = ["first_name",
                  "last_name",
                  "username",
                  "email",
                  "password"]

# Form for login (used for dropdown login)
class AuthenticateDropdownForm(AuthenticationForm):
    def __init__(self, *args, **kwargs):
        error_messages = kwargs.pop("error_messages", {})
        super(AuthenticateDropdownForm, self).__init__(*args, **kwargs)
        for field in self.fields.values():
            for key in error_messages.keys():
                field.error_messages[key] = error_messages[key]

# Form for login (used on home page)
class AuthenticateForm(AuthenticationForm):
    def __init__(self, *args, **kwargs):
        error_messages = kwargs.pop("error_messages", {})
        super(AuthenticateForm, self).__init__(*args, **kwargs)
        for field in self.fields.values():
            for key in error_messages.keys():
                field.error_messages[key] = error_messages[key]

# Form for experiment description selection
class ExpDescriptionSelectForm(forms.Form):
    expDescs = forms.ChoiceField(label="Experiment Descriptions:")

    def __init__(self, *args, **kwargs):
        self.expDescs = kwargs.pop("expDescs", [])
        self.selected = kwargs.pop("selected", "")
        self.selectedIndex = ""
        blank = kwargs.pop("blank", False)
        super(ExpDescriptionSelectForm, self).__init__(*args, **kwargs)

        for i, exp in enumerate(self.expDescs):
           if exp == self.selected:
              self.selectedIndex = exp

        self.fields["expDescs"].choices = [(exp.id, exp.name) for exp in self.expDescs]
        self.fields["expDescs"].initial = self.selectedIndex

        if blank:
            self.fields["expDescs"].choices.insert(0, ("", "None"))

    def getExpDescriptions(self):
        return self.expDescs

# Field for the selection of privacy settings
class PublicChoiceField(forms.ChoiceField):
    def __init__(self, *args, **kwargs):
        CHOICES = [('0', 'Private'), ('1', 'Public')]
        super(PublicChoiceField, self).__init__(*args, choices=CHOICES, **kwargs)

# Form for the display of experiment description information
class ExpDescriptionInfoForm(forms.ModelForm):
    public = PublicChoiceField()

    def __init__(self, *args, **kwargs):
        self.expDescription = kwargs.pop("expDescription", None)
        super(ExpDescriptionInfoForm, self).__init__(*args, **kwargs)
         
        for field in self.fields.values():
            field.widget.attrs["readonly"] = True
        self.fields["public"].widget.attrs["disabled"] = True

        self.fields["name"].initial = self.expDescription.name
        self.fields["description"].initial = self.expDescription.description
        self.fields["public"].initial = self.expDescription.public

    class Meta:
        model = ExpDescription
        fields = ["name", "description", "public"]

# Form for the creation of an experiment
class ExpForm(forms.ModelForm):
  public = PublicChoiceField()

  def __init__(self, *args, **kwargs):
    super(ExpForm, self).__init__(*args, **kwargs)

  class Meta:
    model = Experiment
    fields = ["name", "description", "replications", "public"]

# Form for the selection of an experiment
class ExpSelectForm(forms.Form):
    experiments = forms.ChoiceField(label="Experiment:")
    
    def __init__(self, *args, **kwargs):
        self.experiments = kwargs.pop("experiments", [])
        super(ExpSelectForm, self).__init__(*args, **kwargs)
        self.fields["experiments"].choices = [(exp.id, exp.name) for exp in self.experiments]

    def getExperiments(self):
        return self.experiments

# Form for the display of experiment information
class ExpInfoForm(forms.ModelForm):
  public = PublicChoiceField()

  def __init__(self, *args, **kwargs):
    self.experiment = kwargs.pop("experiment", None)
    super(ExpInfoForm, self).__init__(*args, **kwargs)

    for field in self.fields.values():
        field.widget.attrs["readonly"] = True
    self.fields["public"].widget.attrs["disabled"] = True

    self.fields["name"].initial = self.experiment.name
    self.fields["description"].initial = self.experiment.description
    self.fields["replications"].initial = self.experiment.replications
    self.fields["public"].initial = self.experiment.public

  class Meta:
    model = Experiment
    fields = ["name", "description", "replications", "public"]

# Form for the input of desired factor levels
class FactorForm(forms.Form):
    def __init__(self, *args, **kwargs):
        self.factors = kwargs.pop("factors", [])
        super(FactorForm, self).__init__(*args, **kwargs)

        for i, factor in enumerate(self.factors):
           if factor.data_type == "double":
               self.fields["factor_%s" % i] = RangedFactorField(label=factor.name, factor=factor)

    def getFactors(self):
        return self.factors

# Form for simulation run selection
class SimrunForm(forms.Form):
    simrun = forms.ChoiceField(label="Sumulation Run")

    def __init__(self, *args, **kwargs):
        self.simruns = kwargs.pop("simruns", [])
        super(SimrunForm, self).__init__(*args, **kwargs)

        self.fields["simrun"].choices = [(simrun.id, simrun.id) for simrun in self.simruns]

    def getSimruns(self):
        return self.simruns

# Form for metric selection
class MetricForm(forms.Form):
    metric = forms.ChoiceField(label="Metric")

    def __init__(self, *args, **kwargs):
        self.metrics = kwargs.pop("metrics", [])
        super(MetricForm, self).__init__(*args, **kwargs)

        self.fields["metric"].choices = [(metric.id, metric.name) for metric in self.metrics]

    def getMetrics(self):
        return self.metrics

# Form for design point selection
class DesignpointForm(forms.Form):
    designpoint = forms.ChoiceField(label="Designpoint")

    def __init__(self, *args, **kwargs):
        self.designpoints = kwargs.pop("designpoints", [])
        super(DesignpointForm, self).__init__(*args, **kwargs)

        self.fields["designpoint"].choices = [(designpoint.id, designpoint.id) for designpoint in self.designpoints ]

    def getDesignpoints(self):
        return self.designpoints

# Field for input of ranged values (specifically for factor levels)
class RangedFactorField(forms.CharField):
  def __init__(self, *args, **kwargs):
    self.factor = kwargs.pop("factor", "")
    forms.CharField.__init__(self, *args, **kwargs)
    
    self.title = self.factor.name;
    self.description =  ("Description: " + self.factor.description +
                         "<br />Units: " + self.factor.units +
                         "<br />Type: " + self.factor.data_type +
                         "<br /><br />" +
                           "Enter semi-colon sperated sets of numbers. " +
                           "Each set is of the form start[, end[, step]] " +
                           "Step will default to 1. Examples are: 1, 10, 2 and 1, 10 and 1.")

    self.widget.attrs.update({"class":"has-popover", 
                              "title":self.title,
                              "data-content":self.description,
                              "data-placement":"auto",
                              "data-container":"body"})

  def getFactor(self):
    return self.factor

  def clean(self, value):
    forms.CharField.clean(self, value)
    numExp = "(?:(?:\s*[0-9]*\.[0-9]+\s*)|(?:\s*[0-9]+\s*))"
    regex = re.compile("((?:" + numExp + "," + numExp + "," + numExp + ")" + 
                       "|(?:" + numExp + "," + numExp + ")" + 
                       "|(?:" + numExp + "))+" + 
                       "(?:;|$)")
    matches = regex.findall(value)
    for match in matches:
      entries = match.split(",")
      for i, entry in enumerate(entries):
        entries[i] = float(entry)
      if len(entries) == 3:
        result = list(discreteRange(entries[0], entries[1], entries[2]))
      if len(entries) == 2:
        result = list(discreteRange(entries[0], entries[1]))
      if len(entries) == 1:
        result = [entries[0]]

    if ";".join(matches) != value:
      raise forms.ValidationError("Invalid Expression.")
    return ( self.factor, result )

# Returns a range of values as specified
def discreteRange(start, end = None, step = None):
  if step == None:
    step = 1
  if end == None:
    end = start
    start = 0
  i = 1
  r = start
  while r <= end:
    yield r 
    r = start + i * step
    i += 1
