/**
 * Main JavaScript plotting logic.
 *
 * @author Chris Main
 */

// For demo purposes only
var srun_id = 13;
var exp_id = 1;

var metrics,
xmin,
xmax;

// Is the left or right y-axis set(occupied)?
var lset, rset = false;

// A queue of colors
var colors = queue()('1')('2')('3')('4')('5');

// Draw the interactive plot key
d3.json("/safe/exp/" + exp_id + "/metrics/", function(error, json) {
  if (error) return console.warn(error);
  metrics = json;

  // Add a new key line for each metric, binding its metric_id to id
  var key_items = d3.select("#key") 
  .selectAll("div") 
  .data(metrics) 
  .enter()
  .append("div") 
  .attr("class","key_line") 
  .attr("id",function(d){return d.id});

  // Add a new key square for each metric
  key_items.append("div") 
  .attr("id", function(d){return "key_square_" + d.id}) 
  .attr("class", "key_square");

  // Add a label for each metric
  key_items.append("div")
  .attr("class","key_label") 
  .text(function(d){return d.name});

  // Listen for clicks to turn on/off the time series
  d3.selectAll(".key_line")
  .on("click", getTimeseriesData);
});

// Configure margins
var margin = {top: 10, right: 40, bottom: 100, left: 40},
margin2 = {top: 430, right: 10, bottom: 20, left: 40},
width = 960 - margin.left - margin.right,
height = 500 - margin.top - margin.bottom,
height2 = 500 - margin2.top - margin2.bottom;

// We need relative time, so we choose a random epoch and set our time as an offset
var formatTime = d3.time.format("%H:%M:%S"),
parseSeconds = function(d) { return new Date(2012, 0, 1, 0, 0, d); };

// Configure time scales
var x = d3.time.scale().range([0, width]).domain([0,0]),
x2 = d3.time.scale().range([0, width]).domain([0,0]),
y = d3.scale.linear().range([height, 0]).domain([0,0]),
y_r = d3.scale.linear().range([height, 0]).domain([0,0]),
y2 = d3.scale.linear().range([height2, 0]);
y2_r = d3.scale.linear().range([height2, 0]);

// Configure axis
var xAxis = d3.svg.axis().scale(x).orient("bottom").tickFormat(formatTime),
xAxis2 = d3.svg.axis().scale(x2).orient("bottom").tickFormat(formatTime),
yAxis = d3.svg.axis().scale(y).orient("left");
yAxis_r = d3.svg.axis().scale(y_r).orient("right");

// Configure brush
var brush = d3.svg.brush()
.x(x2)
.on("brush", onBrush);

// Configure line for focus plot
var line = d3.svg.line()
.x(function(d) { return x(d.time); })
.y(function(d) { return y(d.value); });

// Configure line for context plot
var line2 = d3.svg.line()
.x(function(d) { return x2(d.time); })
.y(function(d) { return y2(d.value); });

// Configure line for focus plot
var line3 = d3.svg.line()
.x(function(d) { return x(d.time); })
.y(function(d) { return y_r(d.value); });

// Configure line for context plot
var line4 = d3.svg.line()
.x(function(d) { return x2(d.time); })
.y(function(d) { return y2_r(d.value); });

// Add main container
var svg = d3.select("body").append("svg")
.attr("width", width + margin.left + margin.right)
.attr("height", height + margin.top + margin.bottom + 15);

// Configure main path
svg.append("defs").append("clipPath")
.attr("id", "clip")
.append("rect")
.attr("width", width)
.attr("height", height);

// Focus plot
var focus = svg.append("g")
.attr("transform", "translate(" + margin.left + "," + margin.top + ")");

// Context plot
var context = svg.append("g")
.attr("transform", "translate(" + margin2.left + "," + margin2.top + ")");

// Add x axis to focus
focus.append("g")
.attr("class", "x axis")
.attr("transform", "translate(0," + height + ")")
.call(xAxis);

// Text label for x axis
svg.append("text")
.attr("x", width / 2)
.attr("y", height + margin.bottom + margin2.bottom)
.attr("dy", "0.3em")
.style("text-anchor", "middle")
.text("Time (HH:MM:SS)");

// Add left y axis to focus
focus.append("g")
.attr("class", "y axis")
.call(yAxis);

// Add text label for left y axis
svg.append("text")
.attr("class", "y1 label")
.attr("transform", "rotate(-90)")
.attr("y", 0)
.attr("x", 0 - (height / 2))
.attr("dy", "1em")
.style("text-anchor", "middle")
.text("");

// Add right y axis to focus
focus.append("g")
.attr("class", "y axis r")
.attr("transform", "translate(" + width + ",0)")
.call(yAxis_r);

// Add text label for right y axis
svg.append("text")
.attr("class", "y2 label")
.attr("transform", "rotate(90)")
.attr("y", 0 - width - margin.left - margin.right)
.attr("x", (height / 2))
.attr("dy", "0.8em")
.style("text-anchor", "middle")
.text("");

// Add x axis
context.append("g")
.attr("class", "x axis")
.attr("transform", "translate(0," + height2 + ")")
.call(xAxis2);

// Add brush
context.append("g")
.attr("class", "x brush")
.call(brush)
.selectAll("rect")
.attr("y", -6)
.attr("height", height2 + 7);

/*
 * Draw a timeseries using the provided data and id (metric_id).
 *
 * @param data The data for the timeseries
 * @param id the metric id
 */
function drawTimeseries(data, id, type, name, units) {

  // If both axes are occupied, do nothing
  if (lset && rset) {
    alert("You cannot plot more than two lines.");
    return;
  }

  // Dequeue a new color and make a stroke class using it
  var color = colors();
  var colorStroke = "stroke" + color;

  // Get the next available y-axis, r or l
  var axisLabel = getFreeAxis();

  // Parse data
  data.forEach(function(d) {
    d.time = parseSeconds(+d.time);
    d.value = +d.value;
  });

  // Determine new domain
  // If only one line, domain is of that line.
  // If two lines, domain is the global max and min of the two
  if (rset) {
    var min = d3.min(data, function(d) { return d.time; });
    var max = d3.max(data, function(d) { return d.time; });
    xmin = (min < xmin) ? min : xmin; 
    xmax = (max > xmax) ? max : xmax;
  } else {
    xmin = d3.min(data, function(d) { return d.time; });
    xmax = d3.max(data, function(d) { return d.time; });
  }

  // Set the domain
  x.domain([xmin, xmax]);
  x2.domain(x.domain());


  if (axisLabel == "y_axis_l") {
    y.domain([0, d3.max(data, function(d) { return d.value; })]);
    y2.domain(y.domain());

    // The domain may have changed, redraw old lines
    focus.selectAll(".focusPath").attr("d", line3);
    context.selectAll(".contextPath").attr("d", line4);

    appendPaths(axisLabel, data, id, colorStroke);

    // Redraw left y axis
    focus.select(".y.axis").call(yAxis);

    // Draw label
    d3.select(".y1.label").text(name + " (" + units + ")");
  } else {
    y_r.domain([0, d3.max(data, function(d) { return d.value; })]);
    y2_r.domain(y_r.domain());

    // The domain may have changed, redraw old lines
    focus.selectAll(".focusPath").attr("d", line);
    context.selectAll(".contextPath").attr("d", line2);

    appendPaths(axisLabel, data, id, colorStroke);

    // Redraw right y axis
    focus.select(".y.axis.r").call(yAxis_r);

    // Draw label
    d3.select(".y2.label").text(name + " (" + units + ")");
  }

  // Redraw circles, domain may have changed
  redrawCircles();

  // Draw new circle/points
  drawCircles(data, id, color, axisLabel);

  onBrush();

  // Highlight square
  d3.selectAll("#key_square_" + id)
  .classed("bg" + color, true);

  // Redraw remaining axis
  context.select(".x.axis").call(xAxis);
  context.select(".y.axis").call(yAxis);
  focus.select(".x.axis").call(xAxis);
}

/*
 * Append the lines for the given data to both focus and context.
 *
 * @param whichY Which y-axis is line on, l or r
 * @param data Data for the line
 * @param id Line id binding
 * @param colorClass Stroke color class
 */
function appendPaths(whichY, data, id, colorClass) {
  var focusLine = (whichY == "y_axis_l") ? line : line3;
  var contextLine = (whichY == "y_axis_l") ? line2 : line4;

  focus.append("path")
  .datum(data)
  .attr("d", focusLine)
  .attr("clip-path", "url(#clip)")
  .attr("class", id + "_path focusPath " + colorClass + " " + whichY); 

  context.append("path")
  .datum(data)
  .attr("d", contextLine)
  .attr("class", id + "_path contextPath " + colorClass); 
}

/*
 * Get the next available free axis.
 *
 * @return the label of the axis
 */
function getFreeAxis() {
  if (!lset) {
    lset = true;
    return "y_axis_l"; 
  } else {
    rset = true;
    return "y_axis_r";
  }
}

/*
 * Retrieve data for a timeseries. This is a callback.
 *
 * @param d Data point
 */
function getTimeseriesData(d) {

  // get the id of the current element 
  var id = d3.select(this).attr("id"); 
  // see if we have an associated time series 
  var ts = d3.selectAll("._"+id+"_path"); 
  var tsPoints = d3.selectAll("._"+id+"_points"); 
  if (ts.empty()){
      var dataPath = "/safe/simrun/" + srun_id + "/metric/" + id + "/results/";
    d3.json(dataPath, function(data){ 
      drawTimeseries(data, id, d.type, d.name, d.units);
    }) 
  } else {
    var tsFocusClasses = focus.selectAll("._"+id+"_path").attr("class").split(" ");
    ts.remove();

    parseClasses(tsFocusClasses);

    if (!rset && !lset) {
      resetAllX();
    }

    tsPoints.remove();

    // Redraw circles, domain may have changed
    redrawCircles();

   // brush.clear();
    onBrush();

    d3.selectAll("#key_square_" + id)
    .classed("bg" + colorNumber, false);
  }
}

/*
 * Parse the classes of a path, extracting the color and line type, taking
 * appropriate actions with each.
 *
 * @param tsFocusClasses A split array of strings, each of which are a class
 */
function parseClasses(tsFocusClasses) {
  var patt1 = /stroke[0-9]/;
  var patt2 = /y_axis_l/;
  var patt3 = /y_axis_r/;

  for (i in tsFocusClasses) {
    if (patt1.test(tsFocusClasses[i])) {
      // Get the color id so we can add it back to the queue
      colorNumber = tsFocusClasses[i].charAt(tsFocusClasses[i].length - 1);

      // Enqueue the color
      colors(colorNumber)
    } else if (patt2.test(tsFocusClasses[i])) {
      lset = false;
      y.domain([0,0]);
      y2.domain([0,0]);
      focus.select(".y.axis").call(yAxis);
      d3.select(".y1.label").text("");

      if (rset) {
        rescaleSingleLine()
      }

      focus.select(".focusPath").attr("d", line3);
      context.select(".contextPath").attr("d", line4);
    } else if (patt3.test(tsFocusClasses[i])) {
      rset = false;
      y_r.domain([0,0]);
      y2_r.domain([0,0]);
      focus.select(".y.axis.r").call(yAxis_r);
      d3.select(".y2.label").text("");

      if (lset) {
        rescaleSingleLine()
      }

      focus.select(".focusPath").attr("d", line);
      context.select(".contextPath").attr("d", line2);
    }
  }
}

/*
 * Gets the remaining line on the plot and rescales it along with the axes.
 */
function rescaleSingleLine() {
  data = d3.select(".focusPath").data()[0];
  x.domain(d3.extent(data, function(d) { return d.time; }));
  x2.domain(x.domain());
  focus.select(".x.axis").call(xAxis);
  context.select(".x.axis").call(xAxis2);
}

/*
 * Redraw all points based on new domain (x values).
 */
function redrawCircles() {
  //focus.selectAll("circle")    
  //.attr("cx", function(d) {return x(d.time)});
  
  // Count number of points falling in current domain
  var counterLeftY = 0;
  focus.selectAll("circle.y_axis_l")
  .each(function (d) {if (x(d.time) > 0 && x(d.time) < width) counterLeftY++;});
  
  // Count number of points falling in current domain
  var counterRightY = 0;
  focus.selectAll("circle.y_axis_r")
  .each(function (d) {if (x(d.time) > 0 && x(d.time) < width) counterRightY++;});

  var position = 0;
  var scaleFactorLeft = getScaleFactor(10, counterLeftY);
  var scaleFactorRight = getScaleFactor(10, counterRightY);

  focus.selectAll("circle.y_axis_l")    
  .attr("cx", function(d) {return x(d.time)})
  .attr("cy", function(d) {return y(d.value)})
  .attr("visibility", function(d, i) {
    var t;
    if ((i % scaleFactorLeft == 0) && x(d.time) >= position) {
      t = "visible";
      position = x(d.time) + 5;
      //console.log(x(d.time));
    } else {
      t = "hidden";
    }
    return t;
  });
 
  position = 0;

  focus.selectAll("circle.y_axis_r")    
  .attr("cx", function(d) {return x(d.time)})
  .attr("cy", function(d) {return y_r(d.value)})
  .attr("visibility", function(d, i) {
    var t;
    if ((i % scaleFactorRight == 0) && x(d.time) >= position) {
      t = "visible";
      position = x(d.time) + 5;
    } else {
      t = "hidden";
    }
    return t;
  });
}

/*
 * Reset all x-axes to have zero domain
 */
function resetAllX() {
  x.domain([0,0]);
  x2.domain([0,0]);
  focus.select(".x.axis").call(xAxis);
  context.select(".x.axis").call(xAxis2);
}

/*
 * A simple queue implementation.
 *
 */
function queue() {
  var head, tail, undefined, f;
  return f = function(x) {
    if (x != undefined) {
      tail = tail ? tail.n = {v: x} : head = {v: x};
      x = f;
    } else {
      x = head ? head.v : undefined;
      head = head == tail ? (tail = undefined) : head.n;
    }
    return x;
  };
}

/**
 * Draws the given data points as circles.
 *
 * @param points The points to draw as circles
 */
function drawCircles(points, id, color, axis) {

  var colorClass = "fill" + color;
  var g = focus.append("g")
  .attr("class", "circles_" + axis)
  var position = 0;
  var scaleFactor = getScaleFactor(10, points.length);

  g.selectAll("circle")
  .data(points)
  .enter()
  .append("circle")
  .attr("cx", function(d) {return x(d.time)})
  .attr("cy", function(d) {
    var t = (axis == "y_axis_l") ? y(d.value) : y_r(d.value);
    return t;
  })
  .attr("r",2)
  .attr("clip-path", "url(#clip)")
  .attr("class", id + "_points " + colorClass + " " + axis) 
  .attr("visibility", function(d, i) {
    var t; 
    if ((i % scaleFactor == 0) && x(d.time) >= position) {
      t = "visible";
      position = x(d.time) + 5;
    } else { 
      t = "hidden"; 
    }
    return t;
  })
  .on('mouseover.tooltip', function(d){
    d3.select(this).transition().attr('r', 5)
    d3.select("text." + "tip").remove()
    focus.append('text')
    .text(d.value)
    .attr('x', x(d.time) + 10)
    .attr('y', y(d.value) - 10)
    .attr("class", "tip");
  })
  .on('mouseout.tooltip', function(d){
    d3.select(this).transition().attr('r', 2)
    d3.select("text." + "tip")
    .transition()
    .duration(500)
    .style('opacity',0)
    .attr('transform','translate(10, -10)')
    .remove()
  });

}

/**
 * Gets the scale factor for points on the graph. This number esentially means
 * plot every x-th point.
 *
 * @param pointsPerPixel How many points to plot per pixel (recommended 0.1)
 * @param size The number of points to be plotted
 * @return the scale factor as an integer
 */
function getScaleFactor(pointsPerPixel, size) {
  return Math.ceil((size / width) * pointsPerPixel);
}

/**
 * Callback listener for brush events.
 */
function onBrush() {
  x.domain(brush.empty() ? x2.domain() : brush.extent());
  focus.selectAll("path.y_axis_l").attr("d", line);
  focus.selectAll("path.y_axis_r").attr("d", line3);


  //if (!brush.empty()) {
    redrawCircles();
  //}

  focus.select(".x.axis").call(xAxis);
}
