/**--------------------------------------
 * The base graph object
 * -------------------------------------
 *
 * This is the base graph object. It is expected that all
 * graphs will make use of this object's initialization and
 * ptototype for their functionality.
 */

safe.graph.base = safe_graph_base;
safe.graph.clipPathNumber = 0;

/**
 * Sets up the layout of the graph and adds it to the specified element.
 * @param margins The padding to be placed around the graphing area {top: __, bottom:__, right:__, left:__}
 * @param width The pixel width of the entire graph including graphing area and margins
 * @param selectedCallBack The function to be called when the user clicks on this graph
 *          This value can be "undefined" in which case nothing will happen (will be called "selectedCallBack(this)")
 * @param parentElement The element to which this graph should be added. This can be
 *          a dom element, an html id, or a d3 object.
 *
 * @return The newly created object
 * */

function safe_graph_base(margins, width, height, selectedCallBack, parentElement){
    this.margin = margins;
    this.height = height - this.margin.top - this.margin.bottom;
    this.width = width - this.margin.left - this.margin.right;
    this.selectedCallBack = selectedCallBack;
    this.parentElement = safe.tools.d3Object(parentElement);

    this.xScale = d3.scale.linear().domain([0, 0]).range([0, this.width]);
    this.yScale = d3.scale.linear().domain([0, 0]).range([this.height, 0]);

    this.xAxis = d3.svg.axis().scale(this.xScale).orient("bottom");
    this.yAxis = d3.svg.axis().scale(this.yScale).orient("left");

    this.svg = this.parentElement
        .append("svg")
        .attr("width", this.width + this.margin.left + this.margin.right)
        .attr("height", this.height + this.margin.top + this.margin.bottom)
        .on("click", this.selected.bind(this));

	// Add x axis
	this.svg.append("g")
		.attr("class", "x axis")
		.attr("transform", "translate(" + this.margin.left + "," + (this.height + this.margin.top) + ")")
        .on("click", this.selected.bind(this))
		.call(this.xAxis);

    // Add y axis
    this.svg.append("g")
		.attr("class", "y axis")
		.attr("transform", "translate(" + this.margin.left + "," + this.margin.top + ")")
		.call(this.yAxis);

	// Text label for x axis
	this.xAxisLabel = this.svg
        .append("text")
        .attr("class", "xLabel")
		.attr("x", this.width / 2)
		.attr("y", this.height + this.margin.top)
		.attr("dy", "3em")
		.style("text-anchor", "middle")
		.text("Original x-axis");

	// Add text label for y axis
	this.yAxisLabel = this.svg
        .append("text")
		.attr("class", "yLabel")
		.attr("transform", "rotate(-90)")
		.attr("y", 0)
		.attr("x", 0 - ((this.height) / 2))
		.attr("dy", "1em")
		.style("text-anchor", "middle")
		.text("Original y-axis");

    // Set the cliping area
    this.clipId = "clip_" + safe.graph.clipPathNumber;
    safe.graph.clipPathNumber += 1;

    this.clip = this.svg.append("defs")
        .append("svg:clipPath")
        .attr("id", this.clipId)
        .append("svg:rect")
        .attr("y", this.margin.top)
        .attr("x", this.margin.left)
        .attr("width", this.width)
        .attr("height", this.height);

    return this;
}

/** Called when the user clicks on this graph */
safe_graph_base.prototype.selected = function(){
    if(this.selectedCallBack){ this.selectedCallBack(this) }
}

/** 
 * Sets the label for the x axis of the graph
 * @param label The desired label as a string
 */
safe_graph_base.prototype.setXAxisLabel = function(label){
    this.xAxisString = label;
    this.svg.selectAll(".xLabel").text(label);
}

/** 
 * Sets the label for the y axis of the graph
 * @param label The desired label as a string
 */
safe_graph_base.prototype.setYAxisLabel = function(label){
    this.yAxisString = label;
    this.svg.selectAll(".yLabel").text(label);
}

/** 
 * Returns the current x axis label
 * @return The current x axis label as a string
 */
safe_graph_base.prototype.getXAxisLabel = function(){
    return this.xAxisString;
}

/** 
 * Returns the current y axis label
 * @return The current y axis label as a string
 */
safe_graph_base.prototype.getYAxisLabel = function(){
    return this.yAxisString;
}

/** 
 * Returns the pixel width of the graph including margins
 * @return the pixel width of the graph including margins
 */
safe_graph_base.prototype.getWidth = function(){
    return this.width + this.margin.left + this.margin.right;
}

/** 
 * Returns the pixel height of the graph including margins
 * @return the pixel height of the graph including margins
 */
safe_graph_base.prototype.getHeight = function(){
    return this.height + this.margin.top + this.margin.bottom;
}


/** 
 * Sets the height of the graph (including margins)
 * @param height The pixel height for the graph
 */
safe_graph_base.prototype.setHeight = function(height){
    if(height === undefined){ return }

    this.svg.attr("height", height);
    this.height = height - this.margin.top - this.margin.bottom;
    this.yAxisLabel.attr("x", 0-((this.height) / 2));
	this.xAxisLabel.attr({"x": this.width / 2, "y": this.height + this.margin.top } );
    this.yScale.range([this.height, 0]);
    this.svg.select(".x.axis").attr("transform", "translate(" + this.margin.left + "," + (this.height + this.margin.top) + ")");
    this.svg.select(".y.axis").call(this.yAxis);
    this.clip.attr({"y": this.margin.top, "height": this.height});
}

/** 
 * Sets the width of the graph (including margins)
 * @param width The pixel width for the graph
 */
safe_graph_base.prototype.setWidth = function(width){
    if(width === undefined){ return }
    this.svg.attr("width", width);
    this.width = width - this.margin.left - this.margin.right;
    this.xScale.range([0, this.width]);
	this.xAxisLabel.attr({"x": this.width / 2, "y": this.height + this.margin.top } );
    this.svg.select(".x.axis").call(this.xAxis);
    this.clip.attr({"x": this.margin.left, "width": this.width});
}

/** 
 * Sets the size of the graph (including margins)
 * @param width The pixel width for the graph
 * @param height The pixel height for the graph
 */
safe_graph_base.prototype.setSize = function(width, height){
    this.setWidth(width);
    this.setHeight(height);
}

/**
 * Sets the domain to be used for the x Axis
 * @param domain The domain for the x-axis in the form [xMin, xMax]
 */
safe_graph_base.prototype.setXDomain = function(domain){
    if(domain != undefined && domain.length == 2) {
        if(isNaN(domain[0]))
            domain[0] = 0;
        if(isNaN(domain[1]))
            domain[1] = 0;
        this.xScale.domain(domain) 
    }
    this.svg.select(".x.axis").call(this.xAxis);
}

/**
 * Sets the domain to be used for the y Axis
 * @param domain The domain for the y-axis in the form [yMin, yMax]
 */
safe_graph_base.prototype.setYDomain = function(domain){
    if(domain != undefined && domain.length == 2) {
        if(isNaN(domain[0]))
            domain[0] = 0;
        if(isNaN(domain[1]))
            domain[1] = 0;
        this.yScale.domain(domain) 
    }
    this.svg.select(".y.axis").call(this.yAxis)
}

/** 
 * Sets the domains to be used for this graph
 * Only data points withing these domains will be displayed
 * @param xDomain The domain for the x axis in the form [xMin, xMax]
 * @param yDomain The domain for the y axis in the form [yMin, yMax]
 */
safe_graph_base.prototype.setDomains = function(xDomain, yDomain){
    this.setXDomain(xDomain);
    this.setYDomain(yDomain);
}

/**
 * Returns the current x domain of this graph
 * @return The current x domain of this graph
 */
safe_graph_base.prototype.getXDomain = function(){
    return this.xScale.domain();
}

/**
 * Returns the current y domain of this graph
 * @return The current y domain of this graph
 */
safe_graph_base.prototype.getYDomain = function(){
    return this.yScale.domain();
}

