/**----------------------------------
 * A graph for lineplots
 *----------------------------------
 *
 * This object uses given data with a specified format to create
 * a graph with confidence intervals. The data given should be 
 * of the form [{x:__, highY:__, midY:__, lowY:__}, ...] 
 */


safe.graph.lineplot = safe_graph_lineplot;

// This plot inherits from the cartesian plot
safe_graph_lineplot.prototype = Object.create(safe.graph.cartesian.prototype);
safe_graph_lineplot.prototype.constructor = safe_graph_lineplot;

/**
 * Creates a new lineplot object
 * @param margins The padding to be placed around the graphing area {top: __, bottom:__, right:__, left:__}
 * @param width The pixel width of the entire graph including graphing area and margins
 * @param selectedCallBack The function to be called when the user clicks on this graph
 *          This value can be "undefined" in which case nothing will happen (will be called "selectedCallBack(this)")
 * @param parentElement The element to which this graph should be added. This can be
 *          a dom element, an html id, or a d3 object.
 *
 * @return The newly created object
 * */
function safe_graph_lineplot(margins, width, height, onSelectedCallBack, parentDomId, pad){
    safe.graph.cartesian.apply(this, [margins, width, height, onSelectedCallBack, parentDomId]);

    this.pad = (pad == undefined) ? false : pad;

    this.line = d3.svg.line()
        .x(function (d) { return this.margin.left + this.xScale(d["x"]) }.bind(this))
        .y(function (d) { return this.margin.top + this.yScale(d["midY"]) }.bind(this));

    this.numberFormat = d3.format(".2f");

    return this;
}

/**
 * Displays the specified data set in the form associated with line graphs
 * @param id The string id for the desired data set specified in "addDataSet"
 */
safe_graph_lineplot.prototype.display = function(id){
    if(this.toDisplay[id] != true) { 
        this.svg.selectAll(".dataGroup_" + id).remove();
        this.svg.selectAll(".dataLine_" + id).remove();
        this.svg.selectAll(".dataPath_" + id).remove();
        return;
    }

    //============================================
    groups = this.svg.selectAll(".dataGroup_" + id).data(this.data[id]);
    groups.enter().append("g").attr("class", "dataGroup_" + id);
    groups.exit().remove();

    //points = groups.selectAll(".subPoint").data(["minValue", "lowDevValue", "medianValue", "highDevValue", "maxValue"]);
    points = groups.selectAll(".subPoint").data(function(d){ 
        values =
               [ {"cx": d["x"], "cy": d["lowY"]},
                 {"cx": d["x"], "cy": d["midY"]},
                 {"cx": d["x"], "cy": d["highY"]}] ;
        return values; });

    points.enter().append("circle")
        .attr("class", "subPoint")
        .attr(this.pointAttributes[id])
        .attr("info", function(d) { return this.numberFormat(d.cx) + "<br/>" + this.numberFormat(d.cy) }.bind(this))
        .on("mouseover", this.mouseOverPoint.bind(this))
        .on("mouseout", this.mouseOutPoint.bind(this));
    points
        .attr("clip-path", "url(#" + this.clipId + ")" )
        .attr("cx", function (d) { return this.margin.left + this.xScale(d.cx) }.bind(this))
        .attr("cy", function (d) { return this.margin.top + this.yScale(d.cy) }.bind(this))
    points.exit().remove();

    //============================================
 
    lines = this.svg.selectAll(".dataLine_" + id).data(this.data[id]);
    lines.enter().append("line").attr("class", "dataLine_" + id).attr({"stroke-width":"1", "stroke":this.pointAttributes[id].stroke});
    lines
        .attr("clip-path", "url(#" + this.clipId + ")" )
        .attr("x1", function (d) { return this.margin.left + this.xScale(d["x"]); }.bind(this))
        .attr("x2", function (d) { return this.margin.left + this.xScale(d["x"]); }.bind(this))
        .attr("y1", function (d) { return this.margin.top + this.yScale(d["lowY"]); }.bind(this))
        .attr("y2", function (d) { return this.margin.top + this.yScale(d["highY"]); }.bind(this));
    lines.exit().remove();

    //=============================================

    paths = this.svg.selectAll(".dataPath_" + id).data([this.data[id]]);

    if (this.data[id].length > 1) {
        paths.enter()
            .append("path")
            .attr("class", "dataPath_" + id)
            .attr("stroke", this.pointAttributes[id].stroke)
            .attr("stroke-width", 1);

        paths
            .attr("clip-path", "url(#" + this.clipId + ")" )
            .attr("d", function (d) { return this.line(d) }.bind(this));

        paths.exit().remove();
    }else{
        paths.remove();
    }
}

/** Updates information about maximum and minimum value in the data sets */
safe_graph_lineplot.prototype.updateDataFeatures = function(){
    this.xMin = this.xMax = this.yMin = this.yMax = undefined;
    Object.keys(this.data).forEach(function(id){
        if(this.toDisplay[id] != true){ return }
        xRange = d3.extent(this.data[id], function(d) { return d["x"] });
        this.xMin = (this.xMin == undefined) ? xRange[0] : d3.min([this.xMin, xRange[0]]);
        this.xMax = (this.xMax == undefined) ? xRange[1] : d3.max([this.xMax, xRange[1]]);

        yMin = d3.min(this.data[id], function(d) { return d["lowY"] });
        yMax = d3.max(this.data[id], function(d) { return d["highY"] });
        this.yMin = (this.yMin == undefined) ? yMin : d3.min([this.yMin, yMin]);
        this.yMax = (this.yMax == undefined) ? yMax : d3.max([this.yMax, yMax]);
    }.bind(this));

    this.setDomains([this.xMin, this.xMax], [0, this.yMax]);
}

/**
 * Sets the domain for the x-axis
 * @param domain The domain for the x-axis in the form [xMin, xMax]
 */
safe_graph_lineplot.prototype.setXDomain = function(domain){
    var valLeft, valRight;
    range = parseFloat(domain[1]) - parseFloat(domain[0]);
    if(this.pad && range > 0){
        pad = 0.1 * range;
        valLeft = parseFloat(domain[0]) - pad;
        valRight = parseFloat(domain[1]) + pad;
    }else if (this.pad){
        pad = parseFloat(domain[1]) > 10 ? 10 : this.xMax;
        valLeft = parseFloat(domain[0]) - pad;
        valRight = parseFloat(domain[0]) + pad;
    }else{
        valLeft = parseFloat(domain[0]);
        valRight = parseFloat(domain[1]);
    }

    safe.graph.cartesian.prototype.setXDomain.apply(this, [[valLeft, valRight]]);
}
