/**-------------------------------------
 * A Colored Infobox Object
 * -------------------------------------
 *
 * The prototype of an object which holds lines of information which
 * can be removed by the user and with which a color is associated.
 * An example is a graph key which is the primary use for this object
 * within safe.
 */

safe.infobox.colored = safe_infobox_colored;

/**
 * The constructor of the information object
 *
 * @param removeCallBack The function to be called when an element
 *        is selected for removal. Called as "removeCallBack(key)"
 *        where key is the key given thorugh "addElement"
 * @param parentElement The element to which this infobox should be
 *        added. It can be and element id, dom element, or d3 object.
 */
function safe_infobox_colored(removeCallBack, parentElement){
    this.removeCallBack = removeCallBack;
    this.dParent = safe.tools.d3Object(parentElement);

    this.dSelf = this.dParent
        .append("div")
        .attr("class", "container well col-xs-12")
        .style("width", "100%");

    this.elements = {};
    this.colors = {};
    this.returnAttributes = {};

}

/**
 * Adds an element to the infobox
 * @param key A unique identifier for this element
 * @param elementString The informational string to be displayed
 * @param elementColor The color to be associated with this element
 * @param returnAttributes A value (or object) to be returned along 
 *        with an element's key when the element is removed.
 */
safe_infobox_colored.prototype.addElement = function(key, elementString, elementColor, returnAttributes){
    this.elements[key] = elementString;
    this.colors[key] = elementColor;
    this.returnAttributes[key] = returnAttributes;
    this.updateElements();
}

/**
 * Removes an element from the infobox
 * @param key The key associated with the element through "addElement"
 */
safe_infobox_colored.prototype.removeElement = function(key){
    delete this.elements[key];
    delete this.colors[key];
    delete this.returnAttributes[key];
    this.updateElements();
}

/**
 * Called when an element is selected for removal by the user
 * @param key The key associated with the element through "addElement"
 */
safe_infobox_colored.prototype.removeSelected = function(key){
    this.removeCallBack(this.returnAttributes[key]);
}

/**
 * Updates the elements being displayed in the infobox
 * to those currently associated with it.
 */
safe_infobox_colored.prototype.updateElements = function(){
	this.dSelf.selectAll("div").remove();
	var displayed = this.dSelf.selectAll("div").data(Object.keys(this.elements))
			.enter()
			.append("div")
			.attr("class", "col-xs-12")

    displayed.append("i")
        .attr("class", "glyphicon glyphicon-remove-circle col-xs-1")
        .style("cursor", "pointer")
        .on("click", function(d){ this.removeSelected(d) }.bind(this));

	displayed.append("div")
		.attr("class", "glyphicon glyphicon-check col-xs-1")
		.style("color", function(d){ return this.colors[d] }.bind(this));

	displayed.append("div")
		.attr("class", "key_label")
		.text(function(d) { return this.elements[d] }.bind(this));

    displayed.sort(function(a, b){ return this.elements[a].localeCompare(this.elements[b]) }.bind(this));
}
