/**-------------------------------------
 * A histogram options object
 * -------------------------------------
 *
 * An object prototype which allows the selection of various
 * setting for a currently selected histogram 
 */

safe.selector.histogram_options = safe_selector_histogram_options;

/* Include functionality of the options selector */
safe_selector_histogram_options.prototype = Object.create(safe.selector.options.prototype);
safe_selector_histogram_options.constructor = safe_selector_histogram_options;


/**
 * Creates a new selection object
 * @param titleElement The dom element that holds the graph's title 
 * @param parentElement The dom element, dom id, or d3 object to which this
 *        selector should be added.
 */
function safe_selector_histogram_options(titleElement,  parentElement){
    this.plotList = {};
    safe.selector.options.apply(this, [titleElement, parentElement]);
}

/**
 * Initializes the visual dom elements for this selector
 */
safe_selector_histogram_options.prototype.initComponents = function(){
    safe.selector.options.prototype.initComponents.apply(this);

    this.dSelf.append("label")
        .attr("class", "col-xs-12")
        .text("Bin Number:");

    var binsDropdown = this.dSelf.append("div")
        .attr("class", "col-xs-6")
        .style("padding", "0px")
            .append("select")
            .attr("class", "form-control")
            .on("change", this.binsDropdownChange.bind(this));

    binsDropdown.selectAll("option").data(["sqrt(n)", "log n"])
        .enter()
        .append("option")
        .text(function(d) { return d });

    var binNumber = this.dSelf.append("div")
        .attr("class", "col-xs-6")
        .style("padding", "0px")
            .append("input")
            .attr("class", "form-control")
            .attr("type", "number")
            .attr("min", 1)
            .on("change", this.binNumberChange.bind(this));

    this.dSelf.append("label")
        .attr("class", "col-xs-6")
        .text("Bin Width:");
    var binWidth = this.dSelf.append("div")
        .attr("class", "col-xs-6")
        .style("padding", "0px")
            .append("input")
            .attr("class", "form-control")
            .attr("type", "number")
            .on("change", this.binWidthChange.bind(this));

    this.frequency = false;
    var freqButton = this.dSelf.append("button")
        .attr("class", "col-xs-12 btn btn-default")
        .text("Frequency/Probability")
        .on("click", this.setSelectedFrequency.bind(this)); 

    this.binsDropdown = binsDropdown[0][0];
    this.frequencyButton = freqButton[0][0];
    this.binNumber = binNumber[0][0];
    this.binWidth = binWidth[0][0];
}

/**
 * Sets the label text boxes based on the currently selected graph
 */
safe_selector_histogram_options.prototype.setLabels = function(){
    safe.selector.options.prototype.setLabels.apply(this);
    if(this.frequency === true){ 
        this.selectedGraph.setYAxisLabel("Frequency") 
    }else { 
        this.selectedGraph.setYAxisLabel("Probability") 
    }

    this.yAxisLabel.value = this.selectedGraph.getYAxisLabel();
}
        
/**
 * Sets whether the graph is displaying by frequency or percentage
 */
safe_selector_histogram_options.prototype.setSelectedFrequency = function(){
    this.frequency = !this.frequency;
    if(this.selectedGraph == undefined){ return }
    this.selectedGraph.setFrequency(this.frequency);

    if(this.frequency === true){ 
        this.selectedGraph.setYAxisLabel("Frequency") 
    }else { 
        this.selectedGraph.setYAxisLabel("Probability") 
    }

    this.yAxisLabel.value = this.selectedGraph.getYAxisLabel();
    this.selectedGraph.display();

    d3.values(this.plotList).forEach(function(d){
        d.setFrequency(this.frequency);
        d.setYAxisLabel(this.yAxisLabel.value);
        d.display();
    }.bind(this));
}

/**
 * Adds a graph to the list that is currently displayed.
 * This is required because some option changes must be 
 * reflected accross all displayed graphs
 */
safe_selector_histogram_options.prototype.addGraph = function(id, graph){
    this.plotList[id] = graph;
    if(!this.selectedGraph){ return }

    this.plotList[id].setBinNumber(this.selectedGraph.getBinNumber());
    if(this.frequency === true){ 
        this.plotList[id].setYAxisLabel("Frequency") 
    }else { 
        this.plotList[id].setYAxisLabel("Probability") 
    }
    this.plotList[id].display();
}

/**
 * Called when the bins dropdown changes
 */
safe_selector_histogram_options.prototype.binsDropdownChange = function(){
    if(this.selectedGraph == undefined || this.selectedGraph.getData() == undefined){ return }

    var select = this.binsDropdown;
    var selected = select.options[select.selectedIndex].value;
    if(selected == "sqrt(n)"){
        this.binNumber.value = Math.ceil(Math.sqrt(this.selectedGraph.getData().length));
        this.binNumberChange();
    }else if(selected == "log n"){
        this.binNumber.value = Math.ceil(1 + 3.3 * Math.log(this.selectedGraph.getData().length));     // Sturges' Rule
        this.binNumberChange();
    }
}

/**
 * Sets the currently selected graph
 * @param graph The graph object that has been selected 
 */
safe_selector_histogram_options.prototype.setSelectedGraph = function(graph){
    safe.selector.options.prototype.setSelectedGraph.apply(this, [graph]);
    this.binNumber.value = this.selectedGraph.getBinNumber();
    this.binWidth.value = this.selectedGraph.getBinWidth();
    this.frequency = this.selectedGraph.getFrequency();
}

/**
 * Called when the selected number of bins has changed
 */
safe_selector_histogram_options.prototype.binNumberChange = function(){
    if(this.selectedGraph.getData() == undefined){ return }
    var binNumber = parseInt(this.binNumber.value);
    if(binNumber < 1){ 
        alert("Must have at least one bin.");
        this.binNumber.value = 1;
        binNumber = 1;
    }
    this.selectedGraph.setBinNumber(binNumber);
    this.binNumber.value = this.selectedGraph.getBinNumber();
    this.binWidth.value = this.selectedGraph.getBinWidth();
    this.selectedGraph.display();

    d3.values(this.plotList).forEach(function(d){
        d.setBinNumber(this.selectedGraph.getBinNumber());
        d.display();
    }.bind(this));
}

/**
 * Called when the selected width of the bins has changed
 */
safe_selector_histogram_options.prototype.binWidthChange = function(){
    if(this.selectedGraph.getData() == undefined){ return }
    var binWidth = parseFloat(this.binWidth.value);
    if(binWidth <= 0){ 
        alert("Bin width must be greater than 0.");
        this.binWidth.value = 1;
        binWidth = 1;
    }
    this.selectedGraph.setBinWidth(binWidth);
    this.binNumber.value = this.selectedGraph.getBinNumber();
    this.binWidth.value = this.selectedGraph.getBinWidth();
    this.selectedGraph.display();

    d3.values(this.plotList).forEach(function(d){
        d.setBinNumber(this.selectedGraph.getBinNumber());
        d.display();
    }.bind(this));
}
