/**-------------------------------------
 * A level selector object
 * -------------------------------------
 *
 * An object prototype which allows the selection of factor levels
 * which are stored in the database.
 */
safe.selector.level = safe_selector_level;

/**
 * Creates a new selection object
 * @param exp_id The experiment id with which this selector is associated
 * @param changeCallBack A function to be called when the selection is changed
 * @param parentElement The dom element, dom id, or d3 object to which this
 *        selector should be added.
 */
function safe_selector_level(exp_id, changeCallBack, parentElement){
    this.exp_id = exp_id;
    this.callBack = changeCallBack;
    this.dParent = safe.tools.d3Object(parentElement);
    this.levelList = {};
}

/**
 * Reads data from the database about available factors and their levels 
 * @param callBack A function to be called when the read is complete
 */
safe_selector_level.prototype.readInitialData = function(callBack){
    d3.json("/safe/exp/" + exp_id + "/levels/", function(error, data){
        if(error){ console.warn(error) }

        Object.keys(data).forEach(function(factor_id){ 
            this.levelList[factor_id] = data[factor_id].levels.sort(); 
        }.bind(this));

        if(callBack != undefined){ callBack() }
    }.bind(this));
}

/**
 * Initializes the visual dom elements for this selector
 */
safe_selector_level.prototype.initComponents = function(attributes, styles){
    this.dSelect = this.dParent
        .append("select")
        .attr("class", "form-control")
        .on("change", this.selectChange.bind(this));
    this.eSelect = this.dSelect[0][0];
}

/**
 * Updates the currently displayed levels based on data in the database
 * @param factor_id The id of the factor whose levels should be displayed
 */
safe_selector_level.prototype.updateLevels = function(factor_id){
    levels = this.dSelect.selectAll("option").data(this.levelList[factor_id])
    levels.enter().append("option");
    levels.attr("level", function(d) { return d }).text(function(d) { return d });
    levels.exit().remove();
    this.eSelect.selectedIndex = 0;
}

/**
 * Sets the levels to be associated with a given factor
 * @param factor_id The id of the factor whose levels are being specified
 * @param levels A list of the levels to associate with the factor
 */
safe_selector_level.prototype.setLevels = function(factor_id, levels){
    this.levelList[factor_id] = levels;
}

/**
 * Called when the selected level is changed
 */
safe_selector_level.prototype.selectChange = function(){
    if(this.callBack){ this.callBack(this.getSelectedLevel()) }
}

/**
 * Returns the currently selected level
 * @return The currently selected level
 */
safe_selector_level.prototype.getSelectedLevel= function(){
    return this.eSelect.options[this.eSelect.selectedIndex].getAttribute("level");
}

/**
 * Returns the minimum selectable level for the specified factor
 * @param factor_id The id of the factor to be observed
 * @return The minimum selectable level of the specified factor
 */
safe_selector_level.prototype.getMinLevel = function(factor_id){
    return this.levelList[factor_id][0];
}

/**
 * Returns the maximum selectable level for the specified factor
 * @param factor_id The id of the factor to be observed
 * @return The maximum selectable level of the specified factor
 */
safe_selector_level.prototype.getMaxLevel = function(factor_id){
    return this.levelList[factor_id][this.levelList[factor_id].length - 1 ];
}
