/**-------------------------------------
 * A metric selector object 
 * -------------------------------------
 *
 * An object prototype which allows the selection of metrics
 * which are stored in the database.
 */
safe.selector.metric = safe_selector_metric;

/**
 * Creates a new selection object
 * @param exp_id The experiment id with which this selector is associated
 * @param changeCallBack A function to be called when the selection is changed
 * @param parentElement The dom element, dom id, or d3 object to which this
 *        selector should be added.
 */
function safe_selector_metric(exp_id, changeCallBack, parentElement){
    this.exp_id = exp_id;
    this.callBack = changeCallBack;
    this.dParent = safe.tools.d3Object(parentElement);
    this.metricList = {};
    this.colorList = {};
    this.distinctionId = undefined;
}

/**
 * Reads data from the database about available metrics 
 * @param callBack A function to be called when the read is complete
 */
safe_selector_metric.prototype.readInitialData = function(callBack){
    d3.json("/safe/exp/" + this.exp_id + "/metrics/", function(error, data){
        if(error) { console.warn(error) }

        data.forEach(function(metric){
            this.metricList[metric.id] = metric;
        }.bind(this));
        callBack();
    }.bind(this));
}

/**
 * Initializes the visual dom elements for this selector
 */
safe_selector_metric.prototype.initComponents = function(){
    this.dSelect = this.dParent
        .append("div")
        .attr("class", "container")
        .style("padding", "0px")
        .style("width", "100%");
    this.eSelect = this.dSelect[0][0];

    this.dSelect.append("div").attr("class", "clearfix");
    this.dSelect.append("div").style("height", "0.5cm");

    select_items = this.dSelect.selectAll(".div").data(Object.keys(this.metricList)).enter()
        .append("div")
        .attr("class", "col-xs-12")
        .style({"padding": "0px", "cursor": "pointer"});

    select_items.append("i")
        .attr("class", function(d) { return "glyphicon glyphicon-unchecked select_metric_square_" + d }.bind(this))
        .attr("metric_id", function(d) { return d }.bind(this))
        .on("click", this.selectChange.bind(this));

    select_items.append("label")
        .attr("class", "col-xs-11")
        .attr("metric_id", function(d) {  return d }.bind(this))
        .text(function(d) { return this.metricList[d].name}.bind(this));

    this.dSelect.append("div").attr("class", "clearfix");
    this.dSelect.append("div").style("height", "0.5cm");
}

/**
 * Sets the color to be associated with a given metric and distinctionId
 * @param distinctionId A unique string to be used to distinguish metrics of the same type
 *        that are from different runs/factors/designpoints/ etc.
 * @param metric_id The id of the metric to be assigned
 * @param color The color to be associated with the metric
 */
safe_selector_metric.prototype.setMetricColor = function(distinctionId, metric_id, color){
    if(this.colorList[distinctionId] == undefined) { this.colorList[distinctionId] = {} }
    this.colorList[distinctionId][metric_id] = color;
    this.updateSelectedMetrics();
}

/**
 * Updates the colors associated with the metrics based on the current
 * distinction id.
 */
safe_selector_metric.prototype.updateSelectedMetrics = function(){
    Object.keys(this.metricList).forEach(function(metric_id){
        if(this.colorList[this.distinctionId] != undefined && this.colorList[this.distinctionId][metric_id] != undefined){
            this.dParent.selectAll(".select_metric_square_" + metric_id)
                .classed("glyphicon-unchecked", false)
                .classed("glyphicon-check", true)
                .style("color", this.colorList[this.distinctionId][metric_id]);
        }else{
            this.dParent.selectAll(".select_metric_square_" + metric_id)
                .classed("glyphicon-unchecked", true)
                .classed("glyphicon-check", false)
                .style("color", "black");
        }
    }.bind(this));
}

/**
 * Sets the distinction id to be used currently
 * @param distinctionId The unique string which identifies metrics
 *        which are to be displayed currently.
 */
safe_selector_metric.prototype.setDistinctionId = function(distinctionId){
    this.distinctionId = distinctionId;
    this.updateSelectedMetrics();
}

/**
 * Called when a metric is selected
 */
safe_selector_metric.prototype.selectChange = function(){
    var metric_id = d3.event.target.getAttribute("metric_id");
    this.callBack(this.metricList[metric_id]);
}

/**
 * Returns the metric object associated with the given metric_id
 * @param metric_id The id of the metric to be returned
 * @return The metric object associated with the given metric_id
 */
safe_selector_metric.prototype.getMetric = function(metric_id){
    return this.metricList[metric_id];
}

/**
 * Clears all colors from the metric selector
 * @param distinctionId The unique string which identifies the set
 *      of metric to clear
 */
safe_selector_metric.prototype.clear = function(distinctionId){
    delete this.colorList[distinctionId];
    this.updateSelectedMetrics();
}
