#!/usr/bin/env python

# -*- test-case-name: framework.test.test_client -*-

from twisted.internet.protocol import Protocol, ProcessProtocol, Factory
from twisted.internet.protocol import ClientCreator, ClientFactory
from twisted.protocols.basic import Int32StringReceiver
from twisted.internet import reactor, protocol
from twisted.internet.error import ProcessExitedAlready, ReactorNotRunning
from twisted.python import log
from logging import DEBUG
from pipeProtocol import PipeProtocol
import logging
import sys
import os
import signal
import time
import argparse
import json
import string
import subprocess
            
#Protocol responsible to connect Sim. Client Manager (worker) <---> Exp. Exec. Manager
class SimulationClientProtocol(Int32StringReceiver):
    """Protocol which is responsible for the communication with the the
    Experiment Execution Manager (EEM).
    """

    def __init__(self, factory):
        self.factory=factory
        #dictionary to track child_processes by design_point
        self.processes = {}
        #Switch demuxing all received messages
        self.handlers = {
            "design_point" : self._handleDesignPoint,
            "finished" : self._handleFinished,
            "kill_dp" : self._handleDPKill}


    def stringReceived(self,raw_data):
        """Method invoked when data is received on the network."""
        #Decode json data to a dictionary
        data = json.loads(raw_data)
        
        #Call the correct handler function
        if data['type'] in self.handlers.keys():
            self.handlers[data['type']](data)
        else:
            raise Exception("Invalid message received: "+str(data))

    def connectionMade(self):
        """Method invoked when a connection to the EEM is made.
        When the connection is made it is necessary to register with the EEM 
        identifying the worker machine.
        """
        log.msg("Connection was made", logLevel=DEBUG)
        
        #Register/authenticates client
        self.sendString(json.dumps({"type": "register", "auth_key": self.factory.auth_key}))
        log.msg("Registration Sent", logLevel=DEBUG)
        t1 = time.time()
        p = subprocess.call(str(os.path.abspath(os.path.join(self.factory.sim_path,"waf"))), cwd=os.path.abspath(self.factory.sim_path))
        t2 = time.time()
        print('BUILD TIME = '+str(t2-t1))
        #Ask for new design point for each child
        for i in range(int(self.factory.children)):
            self.sendString(json.dumps({"type": "dp_request", 'id':str(i)}))
            log.msg("Next Simulation Requested", loglevel=DEBUG)

    def _handleDesignPoint(self,msg):
        """Method invoked when a new design point has been received."""
        log.msg("Received Design Point: "+str(msg), logLevel=DEBUG)

        runArg = "--run="+self.factory.sim_script

        for key in msg['input']:
            runArg = runArg+" --"+str(key)+"="+str(msg['input'][key])

        log.msg("RUN ARG: "+runArg)
        #Remove process from previous design_point and add to new design_point
        prev_dp = int(msg['prev_dp_id'])
        if prev_dp != 0:
            for i,x in enumerate(self.processes.get(prev_dp,[])):
                if int(x.id) == int(msg['id']):
                    self.processes[prev_dp].pop(i)
                    break

        p = PipeProtocol(self, msg['id'])
        dp_id = int(msg['dp_id'])
        self.processes[dp_id] = self.processes.get(dp_id, [])
        self.processes[dp_id].append(p)
        
        programPath = str(os.path.abspath(os.path.join(self.factory.sim_path,"waf")))
        reactor.spawnProcess(p,
            programPath,
            ["waf", runArg],
            None,
            path=os.path.abspath(self.factory.sim_path),
            childFDs={0: 'w', 1: 'r', 2: 'r', 3: 'r'})
        print("Design Point Started with pid "+str(p.transport.pid))

    def _handleDPKill(self, msg):
        """Handler for the kill_dp message. It will kill the process group
        toplevel processes (waf) for simulators running the given design_point."""
        dp_id = int(msg['dp_id'])
        for p in self.processes.get(dp_id, []):
            try:
                kill_child_processes(int(p.transport.pid))
                #os.killpg(int(p.transport.pid), signal.SIGTERM)
            except ProcessExitedAlready:
                pass            
            except (ValueError, TypeError):
                #put in place because occasionally p.transport.pid was returning a value other than an int
                pass
        self.processes.pop(dp_id, None)
        
    def _handleFinished(self,data):
        """
        Method invoked when the execution manager notifies us the active
        simulation should be finished. 
        """
        log.msg("Handling FINISHED", logLevel=DEBUG)
        try:
            reactor.stop()
        except ReactorNotRunning:
            #if the reactor isn't running, don't throw an error
            pass

    
    def connectionLost(self,reason):
        """Method invoked when the connection to the EEM is lost."""

        #log.msg("Client connection lost", logLevel=DEBUG)

class SimulationClientFactory(ClientFactory):
    """Factory responsible for the generation of the SimulationClient."""

    protocol = None

    def __init__(self, auth_key, sim_path, sim_script, children):
        self.auth_key=auth_key
        self.sim_path=sim_path
        self.sim_script=sim_script
        self.children = children
    def buildProtocol(self,addr):
        """Method which builds the SimulationClient.

        This should not allow more than one SimulationClient to be instantiated
        at a time, and should raise an exception if a second one is asked to be
        created.

        """
        if not self.protocol:
            self.protocol = SimulationClientProtocol(self)
            return self.protocol
        else:
            raise Exception("Connection already exists")
            return None

    def clientConnectionLost(self,connector,reason):
        """Method invoked when a connection to the EEM is lost."""
        
        try:
            reactor.stop()
        except ReactorNotRunning:
            #if the reactor isn't running, don't throw an error
            pass

    def clientConnectionFailed(self,connector,reason):
        """Method invoked when a connection to the EEM fails."""

        log.msg('Connection failed: %s' % (reason), logLevel=DEBUG)
        try:
            reactor.stop()
        except ReactorNotRunning:
            #if the reactor isn't running, don't throw an error
            pass


#Code obtained from stackoverflow question with url: http://stackoverflow.com/questions/3332043/obtaining-pid-of-child-process            
def kill_child_processes(parent_pid, sig=signal.SIGTERM):
    """Kills a process and all of it's child processes"""
    ps_command = subprocess.Popen("ps -o pid --ppid "+str(parent_pid)+" --noheaders" , shell=True, stdout=subprocess.PIPE)
    ps_output = ps_command.stdout.read()
    retcode = ps_command.wait()
    assert retcode == 0, "ps command returned %d" % retcode
    for pid_str in ([int(parent_pid)] + ps_output.split("\n")[:-1]):
        os.kill(int(pid_str), sig)
           
if __name__ == "__main__":
  #Parse the necessary arguments
  parser = argparse.ArgumentParser(description='Controls the simulator in worker machines')
  parser.add_argument('host', help="Experiment Execution Manager host")
  parser.add_argument('port', help="Experiment Execution Manager port", type=int)
  parser.add_argument('auth_key', help="Authentication key necessary to connect to EEM")
  parser.add_argument('sim_path', help="Simulator folder path")
  parser.add_argument('sim_script', help="Simulator script to execute")
  parser.add_argument('children', help="number of child processes to start on worker", default=1)
  args = parser.parse_args()
  
  log.startLogging(sys.stdout)
  scf = SimulationClientFactory(args.auth_key, args.sim_path, args.sim_script, args.children)

  reactor.connectTCP(args.host,args.port,scf)

  reactor.run()
